# requirements !pip install hubspot3 -q



import pytz
import pandas as pd

import requests
from hubspot3 import Hubspot3



"""

Description for get_deals_all_fiels function

MANDATORY INPUT: 

- API_KEY
- start_date as string "yyyy-mm-dd"
- end_date as string "yyyy-mm-dd"


"""




def get_hubspot_deals_all_fields(API_KEY, start_date, end_date):
    
    from datetime import date as dt, timedelta as td , datetime, timezone

    url = "https://api.hubapi.com/crm/v3/objects/deals"

    
    
    start_date = datetime.strptime(start_date,"%Y-%m-%d")
    
    end_date = datetime.strptime(end_date,"%Y-%m-%d")

    y = 0

    i = 20
    after = '1'
    deals_all_fields_record = []


    while i >0 :
        # miss_tariffa___motivo_ko
        # miss_tariffa___motivo_trash
        # google_ad_click_id
        # contattabile

        if y%500 == 0:
            print("Iteration num.", y)

        y += 1
        
        querystring = {"limit": "100", "paginateAssociations": "false" ,"properties":"tipo_di_ordine,miss_tariffa___motivo_ko,chiave_teamwork,miss_tariffa___motivo_trash,google_ad_click_id,amount_in_home_currency,days_to_close,hs_acv,hs_analytics_source,hs_analytics_source_data_1,hs_analytics_source_data_2,hs_arr,campaign_name,hs_closed_amount,hs_closed_amount_in_home_currency,hs_created_by_user_id,hs_deal_stage_probability,hs_forecast_amount,hs_forecast_probability,hs_is_closed,hs_is_closed_won,hs_lastmodifieddate,hs_likelihood_to_close,hs_line_item_global_term_hs_discount_percentage,hs_line_item_global_term_hs_discount_percentage_enabled,hs_line_item_global_term_hs_recurring_billing_period,hs_line_item_global_term_hs_recurring_billing_period_enabled,hs_line_item_global_term_hs_recurring_billing_start_date,hs_line_item_global_term_hs_recurring_billing_start_date_enabled,hs_line_item_global_term_recurringbillingfrequency,hs_line_item_global_term_recurringbillingfrequency_enabled,hs_manual_forecast_category,hs_merged_object_ids,hs_mrr,hs_next_step,hs_object_id,hs_predicted_amount,hs_predicted_amount_in_home_currency,hs_projected_amount,hs_projected_amount_in_home_currency,hs_tcv,hs_time_in_appointmentscheduled,hs_time_in_closedlost,hs_time_in_closedwon,hs_time_in_contractsent,hs_time_in_decisionmakerboughtin,hs_time_in_presentationscheduled,hs_time_in_qualifiedtobuy,hs_unique_creation_key,hs_updated_by_user_id,hs_user_ids_of_all_owners,hubspot_owner_assigneddate,application_uuid3,dealname,salesforcelastsynctime,amount,dealstage,onboardingfrontlink,pipeline,closedate,onboardingadminlink,createdate,engagements_last_meeting_booked,engagements_last_meeting_booked_campaign,engagements_last_meeting_booked_medium,engagements_last_meeting_booked_source,hs_latest_meeting_activity,hs_sales_email_last_replied,hubspot_owner_id,notes_last_contacted,notes_last_updated,notes_next_activity_date,num_contacted_notes,num_notes,hs_createdate,hubspot_team_id,dealtype,hs_all_owner_ids,description,hs_all_team_ids,hs_all_accessible_team_ids,num_associated_contacts,closed_lost_reason,closed_won_reason,application_uuid2","after":after,
                "archived": "false", "hapikey": API_KEY}

        headers = {'accept': 'application/json'}
        response = requests.request("GET", url, headers=headers, params=querystring)


        dati = response.json()
        try:
            after = dati['paging']['next']['after']
            i = len(dati['paging']['next']['after'])
        except:
            after = ''
            i = 0
        d1 = dati["results"]
        for dd in d1:
            try:
                id = dd['id']
            except :
                id = ''
            try:
                amount_in_home_currency = dd['properties']['amount_in_home_currency']
            except :
                amount_in_home_currency = ''
            try:
                days_to_close = dd['properties']['days_to_close']
            except :
                days_to_close = ''
            try:
                hs_acv = dd['properties']['hs_acv']
            except :
                hs_acv = ''

            try:
                hs_analytics_source = dd['properties']['hs_analytics_source']
            except :
                hs_analytics_source = ''

            try:
                tipo_di_ordine = dd['properties']['tipo_di_ordine']
            except :
                tipo_di_ordine = ''
            try:
                hs_analytics_source_data_1 = dd['properties']['hs_analytics_source_data_1']
            except :
                hs_analytics_source_data_1 = ''
            try:
                hs_analytics_source_data_2 = dd['properties']['hs_analytics_source_data_2']
            except :
                hs_analytics_source_data_2 = ''
            try:
                miss_tariffa___motivo_ko = dd['properties']['miss_tariffa___motivo_ko']
            except :
                miss_tariffa___motivo_ko = ''
            try:
                miss_tariffa___motivo_trash = dd['properties']['miss_tariffa___motivo_trash']
            except :
                miss_tariffa___motivo_trash = ''
            try:
                google_ad_click_id = dd['properties']['google_ad_click_id']
            except :
                google_ad_click_id = ''
            try:
                hs_arr = dd['properties']['hs_arr']
            except :
                hs_arr = ''
            try:
                hs_campaign = dd['properties']['campaign_name']
            except :
                hs_campaign = ''
            try:
                hs_closed_amount = dd['properties']['hs_closed_amount']
            except :
                hs_closed_amount = ''
            try:
                hs_closed_amount_in_home_currency = dd['properties']['hs_closed_amount_in_home_currency']
            except :
                hs_closed_amount_in_home_currency = ''
            try:
                hs_created_by_user_id = dd['properties']['hs_created_by_user_id']
            except :
                hs_created_by_user_id = ''
            try:
                hs_deal_stage_probability = dd['properties']['hs_deal_stage_probability']
            except :
                hs_deal_stage_probability = ''
            try:
                hs_forecast_amount = dd['properties']['hs_forecast_amount']
            except :
                hs_forecast_amount = ''
            try:
                hs_forecast_probability = dd['properties']['hs_forecast_probability']
            except :
                hs_forecast_probability = ''
            try:
                hs_is_closed = dd['properties']['hs_is_closed']
            except :
                hs_is_closed = ''
            try:
                hs_is_closed_won = dd['properties']['hs_is_closed_won']
            except :
                hs_is_closed_won = ''
            try:
                hs_lastmodifieddate = dd['properties']['hs_lastmodifieddate']
            except :
                hs_lastmodifieddate = ''
            try:
                hs_likelihood_to_close = dd['properties']['hs_likelihood_to_close']
            except :
                hs_likelihood_to_close = ''

            try:
                hs_manual_forecast_category = dd['properties']['hs_manual_forecast_category']
            except :
                hs_manual_forecast_category = ''
            try:
                hs_merged_object_ids = dd['properties']['hs_merged_object_ids']
            except :
                hs_merged_object_ids = ''
            try:
                hs_mrr = dd['properties']['hs_mrr']
            except :
                hs_mrr = ''
            try:
                hs_next_step = dd['properties']['hs_next_step']
            except :
                hs_next_step = ''
            try:
                hs_object_id = dd['properties']['hs_object_id']
            except :
                hs_object_id = ''
            try:
                hs_predicted_amount = dd['properties']['hs_predicted_amount']
            except :
                hs_predicted_amount = ''
            try:
                hs_predicted_amount_in_home_currency = dd['properties']['hs_predicted_amount_in_home_currency']
            except :
                hs_predicted_amount_in_home_currency = ''
            try:
                hs_projected_amount = dd['properties']['hs_projected_amount']
            except :
                hs_projected_amount = ''
            try:
                hs_projected_amount_in_home_currency = dd['properties']['hs_projected_amount_in_home_currency']
            except :
                hs_projected_amount_in_home_currency = ''
            try:
                hs_tcv = dd['properties']['hs_tcv']
            except :
                hs_tcv = ''
            try:
                hs_time_in_appointmentscheduled = dd['properties']['hs_time_in_appointmentscheduled']
            except :
                hs_time_in_appointmentscheduled = ''
            try:
                hs_time_in_closedlost = dd['properties']['hs_time_in_closedlost']
            except :
                hs_time_in_closedlost = ''
            try:
                hs_time_in_closedwon = dd['properties']['hs_time_in_closedwon']
            except :
                hs_time_in_closedwon = ''
            try:
                hs_time_in_contractsent = dd['properties']['hs_time_in_contractsent']
            except :
                hs_time_in_contractsent = ''
            try:
                hs_time_in_decisionmakerboughtin = dd['properties']['hs_time_in_decisionmakerboughtin']
            except :
                hs_time_in_decisionmakerboughtin = ''
            try:
                hs_time_in_presentationscheduled = dd['properties']['hs_time_in_presentationscheduled']
            except :
                hs_time_in_presentationscheduled = ''
            try:
                hs_time_in_qualifiedtobuy = dd['properties']['hs_time_in_qualifiedtobuy']
            except :
                hs_time_in_qualifiedtobuy = ''
            try:
                hs_unique_creation_key = dd['properties']['hs_unique_creation_key']
            except :
                hs_unique_creation_key = ''
            try:
                hs_updated_by_user_id = dd['properties']['hs_updated_by_user_id']
            except :
                hs_updated_by_user_id = ''
            try:
                hs_user_ids_of_all_owners = dd['properties']['hs_user_ids_of_all_owners']
            except :
                hs_user_ids_of_all_owners = ''
            try:
                hubspot_owner_assigneddate = dd['properties']['hubspot_owner_assigneddate']
            except :
                hubspot_owner_assigneddate = ''        
            try:
                application_uuid3 = dd['properties']['application_uuid3']
            except :
                application_uuid3 = ''
            try:
                dealname = dd['properties']['dealname']
            except :
                dealname = ''
            try:
                salesforcelastsynctime = dd['properties']['salesforcelastsynctime']
            except :
                salesforcelastsynctime = ''
            try:
                amount = dd['properties']['amount']
            except :
                amount = ''
            try:
                dealstage = dd['properties']['dealstage']
            except :
                dealstage = ''
            try:
                onboardingfrontlink = dd['properties']['onboardingfrontlink']
            except :
                onboardingfrontlink = ''
            try:
                pipeline = dd['properties']['pipeline']
            except :
                pipeline = ''
            try:
                closedate = dd['properties']['closedate']
            except :
                closedate = ''
            try:
                onboardingadminlink = dd['properties']['onboardingadminlink']
            except :
                onboardingadminlink = ''
            try:
                createdate = dd['properties']['createdate']
            except :
                createdate = ''     
            try:
                hs_latest_meeting_activity = dd['properties']['hs_latest_meeting_activity']
            except :
                hs_latest_meeting_activity = ''
            try:
                hs_sales_email_last_replied = dd['properties']['hs_sales_email_last_replied']
            except :
                hs_sales_email_last_replied = ''
            try:
                hubspot_owner_id = dd['properties']['hubspot_owner_id']
            except :
                hubspot_owner_id = ''
            try:
                notes_last_contacted = dd['properties']['notes_last_contacted']
            except :
                notes_last_contacted = ''
            try:
                notes_last_updated = dd['properties']['notes_last_updated']
            except :
                notes_last_updated = ''
            try:
                notes_next_activity_date = dd['properties']['notes_next_activity_date']
            except :
                notes_next_activity_date = ''
            try:
                num_contacted_notes = dd['properties']['num_contacted_notes']
            except :
                num_contacted_notes = ''
            try:
                num_notes = dd['properties']['num_notes']
            except :
                num_notes = ''
            try:
                hs_createdate = dd['properties']['hs_createdate']
            except :
                hs_createdate = ''
            try:
                hubspot_team_id = dd['properties']['hubspot_team_id']
            except :
                hubspot_team_id = ''
            try:
                dealtype = dd['properties']['dealtype']
            except :
                dealtype = ''
            try:
                hs_all_owner_ids = dd['properties']['hs_all_owner_ids']
            except :
                hs_all_owner_ids = ''
            try:
                description = dd['properties']['description']
                description = description.replace(';','',1000)
            except :
                description = ''
            try:
                hs_all_team_ids = dd['properties']['hs_all_team_ids'].replace(';','',1000)
            except :
                hs_all_team_ids = ''
            try:
                hs_all_accessible_team_ids = dd['properties']['hs_all_accessible_team_ids']
            except :
                hs_all_accessible_team_ids = ''
            try:
                num_associated_contacts = dd['properties']['num_associated_contacts']
            except :
                num_associated_contacts = ''
            try:
                closed_lost_reason = dd['properties']['closed_lost_reason']
            except :
                closed_lost_reason = ''
            try:
                closed_won_reason = dd['properties']['closed_won_reason']
            except :
                closed_won_reason = ''
            try:
                chiave_teamwork = dd['properties']['chiave_teamwork']
            except :
                chiave_teamwork = ''

            deals_all_fields_record.append({
                'id' : id ,
                'amount_in_home_currency' : amount_in_home_currency ,
                'days_to_close' : days_to_close ,
                'hs_acv' : hs_acv ,
                'hs_analytics_source' : hs_analytics_source ,
                'hs_analytics_source_data_1' : hs_analytics_source_data_1 ,
                'hs_analytics_source_data_2' : hs_analytics_source_data_2 ,
                'hs_arr' : hs_arr ,
                'hs_campaign' : hs_campaign ,
                'hs_closed_amount' : hs_closed_amount ,
                'hs_closed_amount_in_home_currency' : hs_closed_amount_in_home_currency ,
                'hs_created_by_user_id' : hs_created_by_user_id ,
                'hs_forecast_amount' : hs_forecast_amount ,
                'hs_forecast_probability' : hs_forecast_probability ,
                'hs_is_closed' : hs_is_closed ,
                'hs_is_closed_won' : hs_is_closed_won ,
                'hs_lastmodifieddate' : hs_lastmodifieddate[:10] ,
                'hs_likelihood_to_close' : hs_likelihood_to_close ,
                'hs_manual_forecast_category' : hs_manual_forecast_category ,
                'hs_merged_object_ids' : hs_merged_object_ids ,
                'hs_mrr' : hs_mrr ,
                'hs_next_step' : hs_next_step ,
                'hs_object_id' : hs_object_id ,
                'hs_predicted_amount' : hs_predicted_amount ,
                'hs_predicted_amount_in_home_currency' : hs_predicted_amount_in_home_currency ,
                'hs_projected_amount' : hs_projected_amount ,
                'hs_projected_amount_in_home_currency' : hs_projected_amount_in_home_currency ,
                'hs_tcv' : hs_tcv ,
                'hs_time_in_appointmentscheduled' : hs_time_in_appointmentscheduled ,
                'hs_time_in_closedlost' : hs_time_in_closedlost ,
                'hs_time_in_closedwon' : hs_time_in_closedwon ,
                'hs_time_in_contractsent' : hs_time_in_contractsent ,
                'hs_time_in_decisionmakerboughtin' : hs_time_in_decisionmakerboughtin ,
                'hs_time_in_presentationscheduled' : hs_time_in_presentationscheduled ,
                'hs_time_in_qualifiedtobuy' : hs_time_in_qualifiedtobuy ,
                'hs_unique_creation_key' : hs_unique_creation_key ,
                'hs_updated_by_user_id' : hs_updated_by_user_id ,
                'hs_user_ids_of_all_owners' : hs_user_ids_of_all_owners ,
                'hubspot_owner_assigneddate' : hubspot_owner_assigneddate ,
                'application_uuid3' : application_uuid3 ,
                'dealname' : dealname ,
                'salesforcelastsynctime' : salesforcelastsynctime ,
                'amount' : amount ,
                'dealstage' : dealstage ,
                'onboardingfrontlink' : onboardingfrontlink ,
                'pipeline' : pipeline ,
                'closedate' : closedate ,
                'onboardingadminlink' : onboardingadminlink ,
                'createdate' : createdate ,
                'hs_latest_meeting_activity' : hs_latest_meeting_activity ,
                'hs_sales_email_last_replied' : hs_sales_email_last_replied ,
                'hubspot_owner_id' : hubspot_owner_id ,
                'notes_last_contacted' : notes_last_contacted ,
                'notes_last_updated' : notes_last_updated ,
                'notes_next_activity_date' : notes_next_activity_date ,
                'num_contacted_notes' : num_contacted_notes ,
                'num_notes' : num_notes ,
                'hs_createdate' : hs_createdate ,
                'hubspot_team_id' : hubspot_team_id ,
                'dealtype' : dealtype ,
                'hs_all_owner_ids' : hs_all_owner_ids ,
                'description' : description ,
                'hs_all_team_ids' : hs_all_team_ids ,
                'hs_all_accessible_team_ids' : hs_all_accessible_team_ids ,
                'num_associated_contacts' : num_associated_contacts ,
                'closed_lost_reason' : closed_lost_reason ,
                'closed_won_reason' : closed_won_reason ,
                'miss_tariffa___motivo_ko' : miss_tariffa___motivo_ko ,
                'miss_tariffa___motivo_trash' : miss_tariffa___motivo_trash ,
                'google_ad_click_id' : google_ad_click_id ,
                'tipo_di_ordine' : tipo_di_ordine,
                'chiave_teamwork' : chiave_teamwork 
            })



    deals_all_fields = pd.DataFrame(deals_all_fields_record)


    deals_all_fields['hs_lastmodifieddate'] = pd.to_datetime(deals_all_fields['hs_lastmodifieddate'], format='%Y-%m-%d')

    deals_all_fields = deals_all_fields[(deals_all_fields['hs_lastmodifieddate'] >= start_date) & (deals_all_fields['hs_lastmodifieddate'] <= end_date)]   

    deals_all_fields = deals_all_fields.reset_index(drop = True)
    
    return deals_all_fields




"""

Description for get_hubspot_contact function

MANDATORY INPUT: 

- API_KEY
- start_date string as "yyyy-mm-dd"
- end_date string as "yyyy-mm-dd"



"""



def get_hubspot_contact(API_KEY, start_date, end_date):
    
    from datetime import datetime as dt, timedelta as td, date
    
    
    
    url = "https://api.hubapi.com/crm/v3/objects/contacts"


    utc=pytz.UTC


    i = 20
    after = '1'
    contact_record = []
    def qek(vl1,val2) :
        try :
            if vl1>val2 :
                True
            else :
                False
        except:
            False
    y = 1

    s = requests.Session()


    while i >0 :

        querystring = {"limit": "100", "paginateAssociations": "false" ,"associations":"deals","properties":"contattabile,campaign_content,campaign_medium,campaign_name,campaign_term,source,first_deal_created_date,hs_additional_emails,hs_all_assigned_business_unit_ids,hs_all_contact_vids,hs_analytics_first_touch_converting_campaign,hs_analytics_last_touch_converting_campaign,hs_calculated_form_submissions,hs_calculated_merged_vids,hs_calculated_mobile_number,hs_calculated_phone_number,hs_calculated_phone_number_area_code,hs_calculated_phone_number_country_code,hs_calculated_phone_number_region_code,hs_count_is_unworked,hs_count_is_worked,hs_created_by_conversations,hs_email_domain,hs_email_sends_since_last_engagement,hs_facebook_ad_clicked,hs_first_engagement_object_id,hs_google_click_id,hs_ip_timezone,hs_is_contact,hs_is_unworked,hs_last_sales_activity_date,hs_last_sales_activity_timestamp,hs_lead_status,hs_legal_basis,hs_object_id,hs_predictivescoringtier,hs_sa_first_engagement_date,hs_sa_first_engagement_descr,hs_sa_first_engagement_object_type,hs_sales_email_last_clicked,hs_sales_email_last_opened,hs_searchable_calculated_international_mobile_number,hs_searchable_calculated_international_phone_number,hs_searchable_calculated_mobile_number,hs_searchable_calculated_phone_number,hs_sequences_is_enrolled,hs_user_ids_of_all_owners,hubspot_owner_assigneddate,num_associated_deals,recent_deal_amount,recent_deal_close_date,total_revenue,company,first_conversion_event_name,hs_analytics_first_url,hs_email_delivered,hs_email_optout,currentlyinworkflow,hs_analytics_last_url,hs_email_open,num_conversion_events,website,first_conversion_date,firstname,hs_analytics_num_page_views,hs_analytics_num_visits,recent_conversion_event_name,hs_analytics_first_timestamp,hs_analytics_num_event_completions,hs_social_twitter_clicks,industry,lastname,mobilephone,recent_conversion_date,email,hs_analytics_last_timestamp,hs_email_last_email_name,hs_email_last_send_date,hs_social_facebook_clicks,num_unique_conversion_events,hs_analytics_first_visit_timestamp,hs_analytics_source,hs_email_last_open_date,hs_latest_meeting_activity,hs_sales_email_last_replied,hs_social_linkedin_clicks,hubspot_owner_id,hs_analytics_source_data_1,hs_social_google_plus_clicks,hubspot_team_id,ip_country,hs_all_owner_ids,hs_analytics_last_visit_timestamp,hs_analytics_source_data_2,hs_email_first_send_date,hs_social_num_broadcast_clicks,ip_state,hs_all_team_ids,hs_analytics_first_referrer,hs_email_first_open_date,ip_city,hs_all_accessible_team_ids,hs_analytics_last_referrer,phone,fax,address,hs_analytics_average_page_views,city,hs_analytics_revenue,state,hs_lifecyclestage_lead_date,hs_lifecyclestage_marketingqualifiedlead_date,hs_lifecyclestage_opportunity_date,zip,country,hs_lifecyclestage_salesqualifiedlead_date,jobtitle,hs_lifecyclestage_customer_date,hubspotscore,closedate,hs_lifecyclestage_subscriber_date,hs_lifecyclestage_other_date,lifecyclestage,createdate,lastmodifieddate,days_to_close,associatedcompanyid","after":after,
                    "archived": "false", "hapikey": API_KEY}

        headers = {'accept': 'application/json'}
        response = s.get(url, headers=headers, params=querystring)


        dati = response.json()
        try:
            after = dati['paging']['next']['after']
            i = len(dati['paging']['next']['after'])
        except:
            after = ''
            i = 0
        d1 = dati["results"]
        # campaign_content,campaign_medium,campaign_name,campaign_term
        for dd in d1:
            try:
                upd = dt.strptime(dd['properties']['createdate'][:-1], '%Y-%m-%dT%H:%M:%S.%f')
            except:
                upd = dt.strptime(dd['properties']['createdate'][:-1], '%Y-%m-%dT%H:%M:%S')
            
            begin_date = dt.strptime(start_date, "%Y-%m-%d").date()
            
            start = (date.today() - begin_date).days
            
            begin_date = (dt.today()-td(days=start)).isoformat()
            
            
            try:
                begin_date = dt.strptime(begin_date,'%Y-%m-%dT%H:%M:%S.%f')
            except:
                begin_date = dt.strptime(begin_date,'%Y-%m-%dT%H:%M:%S')
            upd = upd.replace(tzinfo=utc)
            begin_date = begin_date.replace(tzinfo=utc)
            
            
            final_date = dt.strptime(end_date, "%Y-%m-%d").date()

            end = (date.today() - final_date).days
            
            final_date = (dt.today()-td(days=end)).isoformat()
            
            
            try:
                final_date = dt.strptime(final_date,'%Y-%m-%dT%H:%M:%S.%f')
            except:
                final_date = dt.strptime(final_date,'%Y-%m-%dT%H:%M:%S')
            
            
            final_date = final_date.replace(tzinfo=utc)

            
            
            if  upd >= begin_date and upd <= final_date:
                if y%250 == 0:
                    print ("Iteration num.", y)
                y= y+ 1
                try:
                    id = dd['id']
                except :
                    id = ''
                try:
                    source = dd['properties']['source']
                except :
                    source = ''
                try:
                    contattabile = dd['properties']['contattabile']
                except :
                    contattabile = ''
                try:
                    campaign_content = dd['properties']['campaign_content']
                except :
                    campaign_content = ''
                try:
                    campaign_medium = dd['properties']['campaign_medium']
                except :
                    campaign_medium = ''
                try:
                    campaign_name = dd['properties']['campaign_name']
                except :
                    campaign_name = ''
                try:
                    campaign_term = dd['properties']['campaign_term']
                except :
                    campaign_term = ''
                try :
                    deal_id = dd['associations']['deals']['results'][0]['id']
                except :
                    deal_id = ''
                try:
                    first_deal_created_date = dd['properties']['first_deal_created_date']
                except:
                    first_deal_created_date = ''
                try:
                    hs_additional_emails = dd['properties']['hs_additional_emails']
                except:
                    hs_additional_emails = ''
                try:
                    hs_all_assigned_business_unit_ids = dd['properties']['hs_all_assigned_business_unit_ids']
                except:
                    hs_all_assigned_business_unit_ids = ''
                try:
                    hs_all_contact_vids = dd['properties']['hs_all_contact_vids']
                except:
                    hs_all_contact_vids = ''
                try:
                    hs_analytics_first_touch_converting_campaign = dd['properties']['hs_analytics_first_touch_converting_campaign']
                except:
                    hs_analytics_first_touch_converting_campaign = ''
                try:
                    hs_analytics_last_touch_converting_campaign = dd['properties']['hs_analytics_last_touch_converting_campaign']
                except:
                    hs_analytics_last_touch_converting_campaign = ''
                try:
                    hs_calculated_form_submissions = dd['properties']['hs_calculated_form_submissions']
                except:
                    hs_calculated_form_submissions = ''
                try:
                    hs_calculated_merged_vids = dd['properties']['hs_calculated_merged_vids']
                except:
                    hs_calculated_merged_vids = ''
                try:
                    hs_calculated_mobile_number = dd['properties']['hs_calculated_mobile_number']
                except:
                    hs_calculated_mobile_number = ''
                try:
                    hs_calculated_phone_number = dd['properties']['hs_calculated_phone_number']
                except:
                    hs_calculated_phone_number = ''
                try:
                    hs_calculated_phone_number_area_code = dd['properties']['hs_calculated_phone_number_area_code']
                except:
                    hs_calculated_phone_number_area_code = ''
                try:
                    hs_calculated_phone_number_country_code = dd['properties']['hs_calculated_phone_number_country_code']
                except:
                    hs_calculated_phone_number_country_code = ''
                try:
                    hs_calculated_phone_number_region_code = dd['properties']['hs_calculated_phone_number_region_code']
                except:
                    hs_calculated_phone_number_region_code = ''
                try:
                    hs_content_membership_registration_domain_sent_to = dd['properties']['hs_content_membership_registration_domain_sent_to']
                except:
                    hs_content_membership_registration_domain_sent_to = ''
                try:
                    hs_content_membership_registration_email_sent_at = dd['properties']['hs_content_membership_registration_email_sent_at']
                except:
                    hs_content_membership_registration_email_sent_at = ''
                try:
                    hs_count_is_unworked = dd['properties']['hs_count_is_unworked']
                except:
                    hs_count_is_unworked = ''
                try:
                    hs_count_is_worked = dd['properties']['hs_count_is_worked']
                except:
                    hs_count_is_worked = ''
                try:
                    hs_created_by_conversations = dd['properties']['hs_created_by_conversations']
                except:
                    hs_created_by_conversations = ''
                try:
                    hs_email_domain = dd['properties']['hs_email_domain']
                except:
                    hs_email_domain = ''
                try:
                    hs_email_sends_since_last_engagement = dd['properties']['hs_email_sends_since_last_engagement']
                except:
                    hs_email_sends_since_last_engagement = ''
                try:
                    hs_facebook_ad_clicked = dd['properties']['hs_facebook_ad_clicked']
                except:
                    hs_facebook_ad_clicked = ''
                try:
                    hs_first_engagement_object_id = dd['properties']['hs_first_engagement_object_id']
                except:
                    hs_first_engagement_object_id = ''
                try:
                    hs_google_click_id = dd['properties']['hs_google_click_id']
                except:
                    hs_google_click_id = ''
                try:
                    hs_ip_timezone = dd['properties']['hs_ip_timezone']
                except:
                    hs_ip_timezone = ''
                try:
                    hs_is_contact = dd['properties']['hs_is_contact']
                except:
                    hs_is_contact = ''
                try:
                    hs_is_unworked = dd['properties']['hs_is_unworked']
                except:
                    hs_is_unworked = ''
                try:
                    hs_last_sales_activity_date = dd['properties']['hs_last_sales_activity_date']
                except:
                    hs_last_sales_activity_date = ''
                try:
                    hs_last_sales_activity_timestamp = dd['properties']['hs_last_sales_activity_timestamp']
                except:
                    hs_last_sales_activity_timestamp = ''
                try:
                    hs_lead_status = dd['properties']['hs_lead_status']
                except:
                    hs_lead_status = ''
                try:
                    hs_legal_basis = dd['properties']['hs_legal_basis']
                except:
                    hs_legal_basis = ''
                try:
                    hs_object_id = dd['properties']['hs_object_id']
                except:
                    hs_object_id = ''
                try:
                    hs_predictivescoringtier = dd['properties']['hs_predictivescoringtier']
                except:
                    hs_predictivescoringtier = ''
                try:
                    hs_sa_first_engagement_date = dd['properties']['hs_sa_first_engagement_date']
                except:
                    hs_sa_first_engagement_date = ''
                try:
                    hs_sa_first_engagement_descr = dd['properties']['hs_sa_first_engagement_descr']
                except:
                    hs_sa_first_engagement_descr = ''
                try:
                    hs_sa_first_engagement_object_type = dd['properties']['hs_sa_first_engagement_object_type']
                except:
                    hs_sa_first_engagement_object_type = ''
                try:
                    hs_sales_email_last_clicked = dd['properties']['hs_sales_email_last_clicked']
                except:
                    hs_sales_email_last_clicked = ''
                try:
                    hs_sales_email_last_opened = dd['properties']['hs_sales_email_last_opened']
                except:
                    hs_sales_email_last_opened = ''
                try:
                    hs_searchable_calculated_international_mobile_number = dd['properties']['hs_searchable_calculated_international_mobile_number']
                except:
                    hs_searchable_calculated_international_mobile_number = ''
                try:
                    hs_searchable_calculated_international_phone_number = dd['properties']['hs_searchable_calculated_international_phone_number']
                except:
                    hs_searchable_calculated_international_phone_number = ''
                try:
                    hs_searchable_calculated_mobile_number = dd['properties']['hs_searchable_calculated_mobile_number']
                except:
                    hs_searchable_calculated_mobile_number = ''
                try:
                    hs_searchable_calculated_phone_number = dd['properties']['hs_searchable_calculated_phone_number']
                except:
                    hs_searchable_calculated_phone_number = ''
                try:
                    hs_sequences_is_enrolled = dd['properties']['hs_sequences_is_enrolled']
                except:
                    hs_sequences_is_enrolled = ''
                try:
                    hs_time_between_contact_creation_and_deal_close = dd['properties']['hs_time_between_contact_creation_and_deal_close']
                except:
                    hs_time_between_contact_creation_and_deal_close = ''
                try:
                    hs_time_between_contact_creation_and_deal_creation = dd['properties']['hs_time_between_contact_creation_and_deal_creation']
                except:
                    hs_time_between_contact_creation_and_deal_creation = ''
                try:
                    hs_time_to_first_engagement = dd['properties']['hs_time_to_first_engagement']
                except:
                    hs_time_to_first_engagement = ''
                try:
                    hs_time_to_move_from_lead_to_customer = dd['properties']['hs_time_to_move_from_lead_to_customer']
                except:
                    hs_time_to_move_from_lead_to_customer = ''
                try:
                    hs_time_to_move_from_marketingqualifiedlead_to_customer = dd['properties']['hs_time_to_move_from_marketingqualifiedlead_to_customer']
                except:
                    hs_time_to_move_from_marketingqualifiedlead_to_customer = ''
                try:
                    hs_time_to_move_from_opportunity_to_customer = dd['properties']['hs_time_to_move_from_opportunity_to_customer']
                except:
                    hs_time_to_move_from_opportunity_to_customer = ''
                try:
                    hs_time_to_move_from_salesqualifiedlead_to_customer = dd['properties']['hs_time_to_move_from_salesqualifiedlead_to_customer']
                except:
                    hs_time_to_move_from_salesqualifiedlead_to_customer = ''
                try:
                    hs_time_to_move_from_subscriber_to_customer = dd['properties']['hs_time_to_move_from_subscriber_to_customer']
                except:
                    hs_time_to_move_from_subscriber_to_customer = ''
                try:
                    hs_user_ids_of_all_owners = dd['properties']['hs_user_ids_of_all_owners']
                except:
                    hs_user_ids_of_all_owners = ''
                try:
                    hubspot_owner_assigneddate = dd['properties']['hubspot_owner_assigneddate']
                except:
                    hubspot_owner_assigneddate = ''
                try:
                    ip__ecomm_bridge__ecomm_synced = dd['properties']['ip__ecomm_bridge__ecomm_synced']
                except:
                    ip__ecomm_bridge__ecomm_synced = ''
                try:
                    ip__ecomm_bridge__source_app_id = dd['properties']['ip__ecomm_bridge__source_app_id']
                except:
                    ip__ecomm_bridge__source_app_id = ''
                try:
                    ip__ecomm_bridge__source_store_id = dd['properties']['ip__ecomm_bridge__source_store_id']
                except:
                    ip__ecomm_bridge__source_store_id = ''
                try:
                    ip__sync_extension__external_source_account_id = dd['properties']['ip__sync_extension__external_source_account_id']
                except:
                    ip__sync_extension__external_source_account_id = ''
                try:
                    ip__sync_extension__external_source_app_id = dd['properties']['ip__sync_extension__external_source_app_id']
                except:
                    ip__sync_extension__external_source_app_id = ''
                try:
                    ip_country_code = dd['properties']['ip_country_code']
                except:
                    ip_country_code = ''
                try:
                    ip_state_code = dd['properties']['ip_state_code']
                except:
                    ip_state_code = ''
                try:
                    ip_zipcode = dd['properties']['ip_zipcode']
                except:
                    ip_zipcode = ''
                try:
                    num_associated_deals = dd['properties']['num_associated_deals']
                except:
                    num_associated_deals = ''
                try:
                    recent_deal_amount = dd['properties']['recent_deal_amount']
                except:
                    recent_deal_amount = ''
                try:
                    recent_deal_close_date = dd['properties']['recent_deal_close_date']
                except:
                    recent_deal_close_date = ''
                try:
                    total_revenue = dd['properties']['total_revenue']
                except:
                    total_revenue = ''
                try:
                    blog_api_demonstration_blog_subscription = dd['properties']['blog_api_demonstration_blog_subscription']
                except:
                    blog_api_demonstration_blog_subscription = ''
                try:
                    blog_example_blog_10500366470_subscription = dd['properties']['blog_example_blog_10500366470_subscription']
                except:
                    blog_example_blog_10500366470_subscription = ''
                try:
                    company = dd['properties']['company']
                except:
                    company = ''
                try:
                    first_conversion_event_name = dd['properties']['first_conversion_event_name']
                except:
                    first_conversion_event_name = ''
                try:
                    hs_analytics_first_url = dd['properties']['hs_analytics_first_url']
                except:
                    hs_analytics_first_url = ''
                try:
                    hs_email_delivered = dd['properties']['hs_email_delivered']
                except:
                    hs_email_delivered = ''
                try:
                    hs_email_optout = dd['properties']['hs_email_optout']
                except:
                    hs_email_optout = ''
                try:
                    currentlyinworkflow = dd['properties']['currentlyinworkflow']
                except:
                    currentlyinworkflow = ''
                try:
                    hs_analytics_last_url = dd['properties']['hs_analytics_last_url']
                except:
                    hs_analytics_last_url = ''
                try:
                    hs_email_open = dd['properties']['hs_email_open']
                except:
                    hs_email_open = ''
                try:
                    num_conversion_events = dd['properties']['num_conversion_events']
                except:
                    num_conversion_events = ''
                try:
                    website = dd['properties']['website']
                except:
                    website = ''
                try:
                    first_conversion_date = dd['properties']['first_conversion_date']
                except:
                    first_conversion_date = ''
                try:
                    firstname = dd['properties']['firstname']
                except:
                    firstname = ''
                try:
                    hs_analytics_num_page_views = dd['properties']['hs_analytics_num_page_views']
                except:
                    hs_analytics_num_page_views = ''
                try:
                    hs_analytics_num_visits = dd['properties']['hs_analytics_num_visits']
                except:
                    hs_analytics_num_visits = ''
                try:
                    recent_conversion_event_name = dd['properties']['recent_conversion_event_name']
                except:
                    recent_conversion_event_name = ''
                try:
                    hs_analytics_first_timestamp = dd['properties']['hs_analytics_first_timestamp']
                except:
                    hs_analytics_first_timestamp = ''
                try:
                    hs_analytics_num_event_completions = dd['properties']['hs_analytics_num_event_completions']
                except:
                    hs_analytics_num_event_completions = ''
                try:
                    hs_social_twitter_clicks = dd['properties']['hs_social_twitter_clicks']
                except:
                    hs_social_twitter_clicks = ''
                try:
                    industry = dd['properties']['industry']
                except:
                    industry = ''
                try:
                    lastname = dd['properties']['lastname']
                except:
                    lastname = ''
                try:
                    mobilephone = dd['properties']['mobilephone']
                except:
                    mobilephone = ''
                try:
                    recent_conversion_date = dd['properties']['recent_conversion_date']
                except:
                    recent_conversion_date = ''
                try:
                    email = dd['properties']['email']
                except:
                    email = ''
                try:
                    hs_analytics_last_timestamp = dd['properties']['hs_analytics_last_timestamp']
                except:
                    hs_analytics_last_timestamp = ''
                try:
                    hs_email_last_email_name = dd['properties']['hs_email_last_email_name']
                except:
                    hs_email_last_email_name = ''
                try:
                    hs_email_last_send_date = dd['properties']['hs_email_last_send_date']
                except:
                    hs_email_last_send_date = ''
                try:
                    hs_social_facebook_clicks = dd['properties']['hs_social_facebook_clicks']
                except:
                    hs_social_facebook_clicks = ''
                try:
                    num_unique_conversion_events = dd['properties']['num_unique_conversion_events']
                except:
                    num_unique_conversion_events = ''
                try:
                    engagements_last_meeting_booked = dd['properties']['engagements_last_meeting_booked']
                except:
                    engagements_last_meeting_booked = ''
                try:
                    engagements_last_meeting_booked_campaign = dd['properties']['engagements_last_meeting_booked_campaign']
                except:
                    engagements_last_meeting_booked_campaign = ''
                try:
                    engagements_last_meeting_booked_medium = dd['properties']['engagements_last_meeting_booked_medium']
                except:
                    engagements_last_meeting_booked_medium = ''
                try:
                    engagements_last_meeting_booked_source = dd['properties']['engagements_last_meeting_booked_source']
                except:
                    engagements_last_meeting_booked_source = ''
                try:
                    hs_analytics_first_visit_timestamp = dd['properties']['hs_analytics_first_visit_timestamp']
                except:
                    hs_analytics_first_visit_timestamp = ''
                try:
                    hs_analytics_source = dd['properties']['hs_analytics_source']
                except:
                    hs_analytics_source = ''
                try:
                    hs_email_last_open_date = dd['properties']['hs_email_last_open_date']
                except:
                    hs_email_last_open_date = ''
                try:
                    hs_latest_meeting_activity = dd['properties']['hs_latest_meeting_activity']
                except:
                    hs_latest_meeting_activity = ''
                try:
                    hs_sales_email_last_replied = dd['properties']['hs_sales_email_last_replied']
                except:
                    hs_sales_email_last_replied = ''
                try:
                    hs_social_linkedin_clicks = dd['properties']['hs_social_linkedin_clicks']
                except:
                    hs_social_linkedin_clicks = ''
                try:
                    hubspot_owner_id = dd['properties']['hubspot_owner_id']
                except:
                    hubspot_owner_id = ''
                try:
                    notes_last_contacted = dd['properties']['notes_last_contacted']
                except:
                    notes_last_contacted = ''
                try:
                    notes_last_updated = dd['properties']['notes_last_updated']
                except:
                    notes_last_updated = ''
                try:
                    notes_next_activity_date = dd['properties']['notes_next_activity_date']
                except:
                    notes_next_activity_date = ''
                try:
                    num_contacted_notes = dd['properties']['num_contacted_notes']
                except:
                    num_contacted_notes = ''
                try:
                    num_notes = dd['properties']['num_notes']
                except:
                    num_notes = ''
                try:
                    hs_analytics_source_data_1 = dd['properties']['hs_analytics_source_data_1']
                except:
                    hs_analytics_source_data_1 = ''
                try:
                    hs_social_google_plus_clicks = dd['properties']['hs_social_google_plus_clicks']
                except:
                    hs_social_google_plus_clicks = ''
                try:
                    hubspot_team_id = dd['properties']['hubspot_team_id']
                except:
                    hubspot_team_id = ''
                try:
                    ip_country = dd['properties']['ip_country']
                except:
                    ip_country = ''
                try:
                    hs_all_owner_ids = dd['properties']['hs_all_owner_ids']
                except:
                    hs_all_owner_ids = ''
                try:
                    hs_analytics_last_visit_timestamp = dd['properties']['hs_analytics_last_visit_timestamp']
                except:
                    hs_analytics_last_visit_timestamp = ''
                try:
                    hs_analytics_source_data_2 = dd['properties']['hs_analytics_source_data_2']
                except:
                    hs_analytics_source_data_2 = ''
                try:
                    hs_email_first_send_date = dd['properties']['hs_email_first_send_date']
                except:
                    hs_email_first_send_date = ''
                try:
                    hs_social_num_broadcast_clicks = dd['properties']['hs_social_num_broadcast_clicks']
                except:
                    hs_social_num_broadcast_clicks = ''
                try:
                    ip_state = dd['properties']['ip_state']
                except:
                    ip_state = ''
                try:
                    hs_all_team_ids = dd['properties']['hs_all_team_ids']
                except:
                    hs_all_team_ids = ''
                try:
                    hs_analytics_first_referrer = dd['properties']['hs_analytics_first_referrer']
                except:
                    hs_analytics_first_referrer = ''
                try:
                    hs_email_first_open_date = dd['properties']['hs_email_first_open_date']
                except:
                    hs_email_first_open_date = ''
                try:
                    ip_city = dd['properties']['ip_city']
                except:
                    ip_city = ''
                try:
                    hs_all_accessible_team_ids = dd['properties']['hs_all_accessible_team_ids']
                except:
                    hs_all_accessible_team_ids = ''
                try:
                    hs_analytics_last_referrer = dd['properties']['hs_analytics_last_referrer']
                except:
                    hs_analytics_last_referrer = ''
                try:
                    phone = dd['properties']['phone']
                except:
                    phone = ''
                try:
                    fax = dd['properties']['fax']
                except:
                    fax = ''
                try:
                    address = dd['properties']['address']
                except:
                    address = ''
                try:
                    hs_analytics_average_page_views = dd['properties']['hs_analytics_average_page_views']
                except:
                    hs_analytics_average_page_views = ''
                try:
                    city = dd['properties']['city']
                except:
                    city = ''
                try:
                    hs_analytics_revenue = dd['properties']['hs_analytics_revenue']
                except:
                    hs_analytics_revenue = ''
                try:
                    state = dd['properties']['state']
                except:
                    state = ''
                try:
                    hs_lifecyclestage_lead_date = dd['properties']['hs_lifecyclestage_lead_date']
                except:
                    hs_lifecyclestage_lead_date = ''
                try:
                    hs_lifecyclestage_marketingqualifiedlead_date = dd['properties']['hs_lifecyclestage_marketingqualifiedlead_date']
                except:
                    hs_lifecyclestage_marketingqualifiedlead_date = ''
                try:
                    hs_lifecyclestage_opportunity_date = dd['properties']['hs_lifecyclestage_opportunity_date']
                except:
                    hs_lifecyclestage_opportunity_date = ''
                try:
                    zip = dd['properties']['zip']
                except:
                    zip = ''
                try:
                    country = dd['properties']['country']
                except:
                    country = ''
                try:
                    hs_lifecyclestage_salesqualifiedlead_date = dd['properties']['hs_lifecyclestage_salesqualifiedlead_date']
                except:
                    hs_lifecyclestage_salesqualifiedlead_date = ''
                try:
                    jobtitle = dd['properties']['jobtitle']
                except:
                    jobtitle = ''
                try:
                    hs_lifecyclestage_customer_date = dd['properties']['hs_lifecyclestage_customer_date']
                except:
                    hs_lifecyclestage_customer_date = ''
                try:
                    hubspotscore = dd['properties']['hubspotscore']
                except:
                    hubspotscore = ''
                try:
                    closedate = dd['properties']['closedate']
                except:
                    closedate = ''
                try:
                    hs_lifecyclestage_subscriber_date = dd['properties']['hs_lifecyclestage_subscriber_date']
                except:
                    hs_lifecyclestage_subscriber_date = ''
                try:
                    hs_lifecyclestage_other_date = dd['properties']['hs_lifecyclestage_other_date']
                except:
                    hs_lifecyclestage_other_date = ''
                try:
                    lifecyclestage = dd['properties']['lifecyclestage']
                except:
                    lifecyclestage = ''
                try:
                    createdate = dd['properties']['createdate']
                except:
                    createdate = ''
                try:
                    lastmodifieddate = dd['properties']['lastmodifieddate']
                except:
                    lastmodifieddate = ''
                try:
                    associatedcompanyid = dd['properties']['associatedcompanyid']
                except:
                    associatedcompanyid = ''
                try:
                    days_to_close = dd['properties']['days_to_close']
                except:
                    days_to_close = ''
                contact_record.append({
                'id' : id ,

                'first_deal_created_date' : first_deal_created_date,
                'hs_additional_emails' : hs_additional_emails,
                'hs_all_assigned_business_unit_ids' : hs_all_assigned_business_unit_ids,
                'hs_all_contact_vids' : hs_all_contact_vids,
                'irst_touch_converting_campaign' : hs_analytics_first_touch_converting_campaign,
                'last_touch_converting_campaign' : hs_analytics_last_touch_converting_campaign,
                'hs_form_submissions' : hs_calculated_form_submissions,
                'hs_merged_vids' : hs_calculated_merged_vids,
                'hs_mobile_number' : hs_calculated_mobile_number,
                'hs_phone_number' : hs_calculated_phone_number,
                'hs_phone_number_area_code' : hs_calculated_phone_number_area_code,
                'hs_phone_number_country_code' : hs_calculated_phone_number_country_code,
                'hs_phone_number_region_code' : hs_calculated_phone_number_region_code,
                'hs_content_domain_sent_to' : hs_content_membership_registration_domain_sent_to,
                'hs_content_email_sent_at' : hs_content_membership_registration_email_sent_at,
                'hs_count_is_unworked' : hs_count_is_unworked,
                'hs_count_is_worked' : hs_count_is_worked,
                'hs_created_by_conversations' : hs_created_by_conversations,
                'hs_email_domain' : hs_email_domain,
                'hs_email_sends_since_last_engagement' : hs_email_sends_since_last_engagement,
                'hs_facebook_ad_clicked' : hs_facebook_ad_clicked,
                'hs_first_engagement_object_id' : hs_first_engagement_object_id,
                'hs_google_click_id' : hs_google_click_id,
                'hs_ip_timezone' : hs_ip_timezone,
                'hs_is_contact' : hs_is_contact,
                'hs_is_unworked' : hs_is_unworked,
                'hs_last_sales_activity_date' : hs_last_sales_activity_date,
                'hs_last_sales_activity_timestamp' : hs_last_sales_activity_timestamp,
                'hs_lead_status' : hs_lead_status,
                'hs_legal_basis' : hs_legal_basis,
                'hs_object_id' : hs_object_id,
                'hs_predictivescoringtier' : hs_predictivescoringtier,
                'hs_sa_first_engagement_date' : hs_sa_first_engagement_date,
                'hs_sa_first_engagement_descr' : hs_sa_first_engagement_descr,
                'hs_sa_first_engagement_object_type' : hs_sa_first_engagement_object_type,
                'hs_sales_email_last_clicked' : hs_sales_email_last_clicked,
                'hs_sales_email_last_opened' : hs_sales_email_last_opened,
                'hs_searchable_calculated_mobile_number' : hs_searchable_calculated_mobile_number,
                'hs_searchable_calculated_phone_number' : hs_searchable_calculated_phone_number,
                'hs_sequences_is_enrolled' : hs_sequences_is_enrolled,
                'hs_time_deal_creation' : hs_time_between_contact_creation_and_deal_creation,
                'hs_time_to_first_engagement' : hs_time_to_first_engagement,
                'hs_time_lead_to_customer' : hs_time_to_move_from_lead_to_customer,
                'hs_time_to_customer' : hs_time_to_move_from_marketingqualifiedlead_to_customer,
                'hs_user_ids_of_all_owners' : hs_user_ids_of_all_owners,
                'hubspot_owner_assigneddate' : hubspot_owner_assigneddate,
                'num_associated_deals' : num_associated_deals,
                'recent_deal_amount' : recent_deal_amount,
                'recent_deal_close_date' : recent_deal_close_date,
                'total_revenue' : total_revenue,
                'blog_api' : blog_api_demonstration_blog_subscription,
                'blog_example_' : blog_example_blog_10500366470_subscription,
                'company' : company,
                'first_conversion_event_name' : first_conversion_event_name,
                'hs_analytics_first_url' : hs_analytics_first_url,
                'hs_email_delivered' : hs_email_delivered,
                'hs_email_optout' : hs_email_optout,
                'currentlyinworkflow' : currentlyinworkflow,
                'hs_analytics_last_url' : hs_analytics_last_url,
                'hs_email_open' : hs_email_open,
                'num_conversion_events' : num_conversion_events,
                'website' : website,
                'first_conversion_date' : first_conversion_date,
                'firstname' : firstname,
                'hs_analytics_num_page_views' : hs_analytics_num_page_views,
                'hs_analytics_num_visits' : hs_analytics_num_visits,
                'recent_conversion_event_name' : recent_conversion_event_name,
                'hs_analytics_first_timestamp' : hs_analytics_first_timestamp,
                'hs_analytics_num_event_completions' : hs_analytics_num_event_completions,
                'hs_social_twitter_clicks' : hs_social_twitter_clicks,
                'industry' : industry,
                'lastname' : lastname,
                'mobilephone' : mobilephone,
                'recent_conversion_date' : recent_conversion_date,
                'email' : email,
                'hs_analytics_last_timestamp' : hs_analytics_last_timestamp,
                'hs_email_last_email_name' : hs_email_last_email_name,
                'hs_email_last_send_date' : hs_email_last_send_date,
                'hs_social_facebook_clicks' : hs_social_facebook_clicks,
                'num_unique_conversion_events' : num_unique_conversion_events,
                'hs_analytics_source' : hs_analytics_source,
                'hs_email_last_open_date' : hs_email_last_open_date,
                'hs_latest_meeting_activity' : hs_latest_meeting_activity,
                'hs_sales_email_last_replied' : hs_sales_email_last_replied,
                'hs_social_linkedin_clicks' : hs_social_linkedin_clicks,
                'hubspot_owner_id' : hubspot_owner_id,
                'notes_last_contacted' : notes_last_contacted,
                'notes_last_updated' : notes_last_updated,
                'notes_next_activity_date' : notes_next_activity_date,
                'num_contacted_notes' : num_contacted_notes,
                'num_notes' : num_notes,
                'hs_analytics_source_data_1' : hs_analytics_source_data_1,
                'hs_social_google_plus_clicks' : hs_social_google_plus_clicks,
                'hubspot_team_id' : hubspot_team_id,
                'ip_country' : ip_country,
                'hs_all_owner_ids' : hs_all_owner_ids,
                'hs_analytics_last_visit_timestamp' : hs_analytics_last_visit_timestamp,
                'hs_analytics_source_data_2' : hs_analytics_source_data_2,
                'hs_email_first_send_date' : hs_email_first_send_date,
                'hs_social_num_broadcast_clicks' : hs_social_num_broadcast_clicks,
                'ip_state' : ip_state,
                'hs_all_team_ids' : hs_all_team_ids,
                'hs_analytics_first_referrer' : hs_analytics_first_referrer,
                'hs_email_first_open_date' : hs_email_first_open_date,
                'ip_city' : ip_city,
                'hs_all_accessible_team_ids' : hs_all_accessible_team_ids,
                'hs_analytics_last_referrer' : hs_analytics_last_referrer,
                'phone' : phone,
                'fax' : fax,
                'address' : address,
                'hs_analytics_average_page_views' : hs_analytics_average_page_views,
                'city' : city,
                'hs_analytics_revenue' : hs_analytics_revenue,
                'state' : state,
                'hs_lifecyclestage_lead_date' : hs_lifecyclestage_lead_date,
                'hs_lifecyclestage_marketingqualifiedlead_date' : hs_lifecyclestage_marketingqualifiedlead_date,
                'hs_lifecyclestage_opportunity_date' : hs_lifecyclestage_opportunity_date,
                'zip' : zip,
                'country' : country,
                'hs_lifecyclestage_salesqualifiedlead_date' : hs_lifecyclestage_salesqualifiedlead_date,
                'jobtitle' : jobtitle,
                'hs_lifecyclestage_customer_date' : hs_lifecyclestage_customer_date,
                'hubspotscore' : hubspotscore,
                'closedate' : closedate,
                'hs_lifecyclestage_subscriber_date' : hs_lifecyclestage_subscriber_date,
                'hs_lifecyclestage_other_date' : hs_lifecyclestage_other_date,
                'lifecyclestage' : lifecyclestage,
                'createdate' : createdate,
                'lastmodifieddate' : lastmodifieddate,
                'associatedcompanyid' : associatedcompanyid,
                'days_to_close' : days_to_close,
                'deal_id' : deal_id,
                'source'  : source,
                'campaign_content' : campaign_content,
                'campaign_medium' : campaign_medium,
                'campaign_name' : campaign_name,
                'campaign_term' : campaign_term,
                'contattabile' : contattabile

                })


    contact = pd.DataFrame(contact_record)
    

    return contact








"""

Description for get_hubspot_pipeline function

MANDATORY INPUT: API_KEY



"""

def get_hubspot_pipelines(API_KEY):                    
    
    from datetime import date as dt, timedelta as td , datetime, timezone


    client = Hubspot3(api_key=API_KEY)

    pipelines_client = client.crm_pipelines.get_all()

    pipeline_dati = []

    for d in pipelines_client:
        for s in d['stages']:
            pipeline_dati.append({
                "label" : d['label'],
                "active" : d['active'],
                "object_type" : d['objectType'],
                "pipeline_id" : d['pipelineId'],
                "created_at" : d['createdAt'],
                "updatet_at" : d['updatedAt'],
                "stage_name" : s['label'],
                "stage_id" : s['stageId'],
                "is_close" : s['metadata']['isClosed']

            })


    pipelines = pd.DataFrame(pipeline_dati)

    pipelines["created_at"] = pd.to_datetime(pipelines["created_at"], unit='ms')

    pipelines["updatet_at"] = pd.to_datetime(pipelines["updatet_at"], unit='ms')

    return pipelines
        



"""

Description for get_hubspot_users function

MANDATORY INPUT: API_KEY


"""

def get_hubspot_users(API_KEY):

    from datetime import date as dt, timedelta as td , datetime, timezone


    url = f"https://api.hubapi.com/owners/v2/owners?hapikey={API_KEY}"

    r= requests.get(url = url)
    dati = r.json()
    users_records = []
    for d in list(dati):
        try:
            users_records.append({
                "portalid" : d["portalId"],
                "ownerid" : d["ownerId"],
                "type" : d["type"],
                "firstname" : d["firstName"],
                "lastname" : d["lastName"],
                "email" : d["email"],
                "is_active" : d["isActive"]
            })

        except:
            pass

    users = pd.DataFrame(users_records)

    return users


import requests
import yaml
import os
import pandas as pd
import numpy as np
import logging
import getpass
import json
import uuid
from askdata.insight import Insight
from askdata.channel import Channel
from askdata.catalog import Catalog
from askdata.dataset import Dataset
from askdata.insight_definition import Insight_Definition
from askdata.security import SignUp
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from typing import TYPE_CHECKING
from IPython.core.display import display, HTML

if TYPE_CHECKING:
    from askdata.askdata_client import Askdata

_LOG_FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s() ] - %(asctime)s --> %(message)s"
g_logger = logging.getLogger()
logging.basicConfig(format=_LOG_FORMAT)
g_logger.setLevel(logging.INFO)

root_dir = os.path.abspath(os.path.dirname(__file__))
# retrieving base url
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to Python the dictionary format
    url_list = yaml.load(file, Loader=yaml.FullLoader)


class Agent(Insight, Channel, Catalog, Dataset):
    '''
    Agent Object
    '''

    def __init__(self, askdata: 'Askdata', slug='', agent_name='', agent_id=''):

        self.username = askdata.username
        self.userid = askdata.userid
        self._domainlogin = askdata._domainlogin
        self._env = askdata._env
        self._token = askdata._token
        self.df_agents = askdata.agents_dataframe()

        self._headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        if self._env == 'dev':
            self.smart_insight_url = url_list['BASE_URL_INSIGHT_DEV']
        if self._env == 'qa':
            self.smart_insight_url = url_list['BASE_URL_INSIGHT_QA']
        if self._env == 'prod':
            self.smart_insight_url = url_list['BASE_URL_INSIGHT_PROD']

        try:
            if slug != '':
                agent = self.df_agents[self.df_agents['slug'] == slug.lower()]
            elif agent_id != '':
                agent = self.df_agents[self.df_agents['id'] == agent_id]
            else:
                agent = self.df_agents[self.df_agents['name'] == agent_name]

            self._agentId = agent.iloc[0]['id']
            self._domain = agent.iloc[0]['domain']
            self._language = agent.iloc[0]['language']
            self._agent_name = agent.iloc[0]['name']
            if(slug!=''):
                self._slug=slug
            else:
                self._slug = agent.iloc[0]['slug']

        except Exception as ex:
            raise NameError('Agent slug/name/id not exsist or not insert')

        Insight.__init__(self, self._env, self._token)
        Channel.__init__(self, self._env, self._token, self._agentId, self._domain)
        Catalog.__init__(self, self._env, self._token)
        #Dataset.__init__(self, self._env, self._token, self._agentId)

    def __str__(self):
        return '{}'.format(self._agentId)

    def Dataset(self, dataset_slug):
        return Dataset(self._env, self._token, self._agentId, dataset_slug)



    def switch_agent(self):

        data = {
            "agent_id": self._agentId
        }

        if self._env == 'dev':
            self._base_url = url_list['BASE_URL_FEED_DEV']
        if self._env == 'qa':
            self._base_url = url_list['BASE_URL_FEED_QA']
        if self._env == 'prod':
            self._base_url = url_list['BASE_URL_FEED_PROD']

        if self._env == 'dev':
            self._base_url_ch = url_list['BASE_URL_FEED_DEV']
        if self._env == 'qa':
            self._base_url_ch = url_list['BASE_URL_FEED_QA']
        if self._env == 'prod':
            self._base_url_ch = url_list['BASE_URL_FEED_PROD']

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = self._base_url + '/' + self._domain + '/agent/switch'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()

        return r

    def ask(self, text, payload=''):

        data = {
            "text": text,
            "payload": payload
        }

        if self._env == 'dev':
            request_agent_url = url_list['BASE_URL_FEED_DEV'] + '/' + self._domain + '/agent/' + self._agentId + '/'
        if self._env == 'qa':
            request_agent_url = url_list['BASE_URL_FEED_QA'] + '/' + self._domain + '/agent/' + self._agentId + '/'
        if self._env == 'prod':
            request_agent_url = url_list['BASE_URL_FEED_PROD'] + '/' + self._domain + '/agent/' + self._agentId + '/'

        response = requests.post(url=request_agent_url, headers=self._headers, json=data)
        response.raise_for_status()
        r = response.json()
        # dataframe creation
        df = pd.DataFrame(np.array(r[0]['attachment']['body'][0]['details']['rows']),
                          columns=r[0]['attachment']['body'][0]['details']['columns'])

        return df

    def ask_as_json(self, text, payload=''):

        data = {
            "text": text,
            "payload": payload
        }

        if self._env == 'dev':
            request_agent_url = url_list['BASE_URL_FEED_DEV'] + '/' + self._domain + '/agent/' + self._agentId + '/'
        if self._env == 'qa':
            request_agent_url = url_list['BASE_URL_FEED_QA'] + '/' + self._domain + '/agent/' + self._agentId + '/'
        if self._env == 'prod':
            request_agent_url = url_list['BASE_URL_FEED_PROD'] + '/' + self._domain + '/agent/' + self._agentId + '/'

        response = requests.post(url=request_agent_url, headers=self._headers, json=data)
        response.raise_for_status()
        r = response.json()

        return r

    def set_current_dataset(self, slug):
        """
        set in the agent object the properties of specific dataset

        :param slug: str, identification of the dataset
        :return: None
        """
        self._get_info_dataset_by_slug(slug)
        return self


    def query(self, dataset_slug="", fields:list=[], filters=None, sorted_by=None,  dataset_id="", limit=50, offset=0, pivot=[]):
        """
        Pivot columns MUST be contained in field list
        """
        if(fields==[]):
            print("SELECT FIELDS")
            return

        if sorted_by is None:
            sorted_by = []

        if filters is None:
            filters = []

        if(dataset_id=="" and dataset_slug!=""):
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            url = self._base_url_askdata + '/smartbot/agents/agentslug/' + self._slug + '/datasetslug/' + dataset_slug

            headers = {"Authorization": "Bearer" + " " + self._token}
            response = s.get(url=url, headers=headers)
            response.raise_for_status()
            r = response.json()

            datasetId = r["datasetId"]

        elif(dataset_id!=""):
            datasetId=dataset_id
        else:
            print("SPECIFY A DATASET ID OR SLUG")
            return

        body = {
            "agentId": self._agentId,
            "datasetId": datasetId,
            "fields": fields,
            "sortedBy": sorted_by,
            "filters": filters,
            "limit": limit,
            "offset": offset,
            "pivot": pivot
        }

        url = self.smart_insight_url + "/composed_queries/sdk"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=url, json=body, headers=headers)
        response.raise_for_status()

        qc_id = response.json()["id"]

        preview_url = self.smart_insight_url + "/composed_queries/" + qc_id + "/preview?agentId="+self._agentId

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=preview_url, json={}, headers=headers)
        response.raise_for_status()

        r = response.json()
        codes = []
        columns = []
        for col in r["schema"]:
            codes.append(col["code"])
            if col["name"] != '':
                columns.append(col["name"])
            else:
                columns.append(col["code"])
        df = pd.DataFrame(columns=codes, data=r["data"])
        df.columns = columns
        return df


    def update_dataset_name(self, dataset_slug, dataset_name):

        body = {"name": dataset_name}

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartdataset/datasets/' + dataset_slug + '/sdk'
        logging.info("AUTH URL {}".format(authentication_url))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        response = s.put(url=authentication_url, json=body, headers=headers)
        response.raise_for_status()

    def create_parquet_dataset(self, agent_slug, dataset_slug, file_path):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartbot/agents/' + agent_slug + '/datasets/' + dataset_slug + '/parquet'
        logging.info("AUTH URL {}".format(authentication_url))
        file = {'file': open(file_path, 'rb')}
        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=authentication_url, files=file, headers=headers)
        response.raise_for_status()
        r = response.json()


    def update_parquet_dataset(self, agent_slug, dataset_id, file_path, strategy):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartbot/agents/' + agent_slug + '/datasets/' + dataset_id + '/parquet?strategy=' + strategy
        logging.info("AUTH URL {}".format(authentication_url))
        file = {'file': open(file_path, 'rb')}
        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.put(url=authentication_url, files=file, headers=headers)
        response.raise_for_status()
        r = response.json()


    def get_dataset_by_slug(self, agent_slug, slug: str):
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        try:
            authentication_url = self._base_url_askdata + '/smartbot/agents/agentslug/' + agent_slug + '/datasetslug/' + slug
            logging.info("AUTH URL {}".format(authentication_url))

            headers = {"Authorization": "Bearer" + " " + self._token}
            response = s.get(url=authentication_url, headers=headers)
            response.raise_for_status()
            r = response.json()
            return r["datasetId"]
        except:
            return None

    def __update_dataset_icon(self, dataset_id, icon_url):

        url = self._base_url_askdata + "/smartdataset/datasets/" + dataset_id + "/settings"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        logging.info("AUTH URL {}".format(url))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        response = s.get(url=url, headers=headers)
        response.raise_for_status()

        settings = response.json()

        settings["icon"] = icon_url

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        logging.info("AUTH URL {}".format(url))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        response = s.put(url=url, json=settings, headers=headers)
        response.raise_for_status()

    def create_dataset(self, dataframe: pd.DataFrame, dataset_name: str, slug: str, icon_url=None,
                       settings: dict = None):

        for col in dataframe.columns:
            if (dataframe[col].dtypes == "datetime64[ns]"):
                dataframe[col] = dataframe[col].astype("str")

        body = {"label": dataset_name, "rows": dataframe.to_dict(orient="record")}

        settings_list = []
        if (settings != None):
            for key in settings.keys():
                settings[key]["column_name"] = key
                settings_list.append(settings[key])
            body["settings"] = settings_list

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self._base_url_askdata + '/smartbot/agents/' + self._agentId + '/datasets/' + slug + '/sdk'
        logging.info("AUTH URL {}".format(url))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        response = s.post(url=url, json=body, headers=headers)
        response.raise_for_status()

        dataset_id = response.json()["id"]

        if (icon_url != None):
            self.__update_dataset_icon(dataset_id, icon_url)

    def update_dataset(self, dataframe: pd.DataFrame, dataset_name: str, slug: str, icon_url=None,
                       settings: dict = None):

        for col in dataframe.columns:
            if (dataframe[col].dtypes == "datetime64[ns]"):
                dataframe[col] = dataframe[col].astype("str")

        body = {"label": dataset_name, "rows": dataframe.to_dict(orient="records")}

        settings_list = []
        if (settings != None):
            for key in settings.keys():
                settings[key]["column_name"] = key
                settings_list.append(settings[key])
            body["settings"] = settings_list

        s = requests.Session()

        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self._base_url_askdata + '/smartbot/agents/' + self._agentId + '/datasets/' + slug + '/sdk'
        logging.info("AUTH URL {}".format(url))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        response = s.put(url=url, json=body, headers=headers)

        response.raise_for_status()
        dataset_id = response.json()["id"]

        if (icon_url != None):
            self.__update_dataset_icon(dataset_id, icon_url)


    def check_dataset_columns(self, dataframe: pd.DataFrame, dataset_id):

        body = dataframe.to_dict(orient="record")

        url = self._base_url_askdata + '/smartdataset/datasets/'+dataset_id+'/sdk/check'

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        logging.info("AUTH URL {}".format(url))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        response = s.post(url=url, json=body, headers=headers)
        response.raise_for_status()
        r = response.json()

        if(r["hasDifference"]):
            for differece in r["differences"]:
                logging.info("Column {} is being {}".format(differece["columnName"], differece["differenceType"]))
        return r["hasDifference"]

    '''La prima chiave è la stringa colonna del dataframe. Per ogni chiave ho dei setting:
    name stringa
    type(dimension o measure)
    isDate booleano
    synonyms: [] array di stringhe
    dateFormat(pattern della data)
    numberFormat
    defaultAggregation(sum, avg..)

    {
        "column_name": {
            "name": "column_name",
            "type": "Dimension",
            "isDate": False,
            "synonyms": [],
            "dateFormat",
            "numberFormat": "",
            "defaultAggregation": "sum"
        }
    }
    '''
    def create_or_replace_dataset(self, dataframe: pd.DataFrame, dataset_name: str, slug: str, icon_url=None,
                                  settings: dict = None):
        # Check if dataset exists
        dataset_id = self.get_dataset_by_slug(self._slug, slug)

        dataframe = dataframe.replace({np.nan: None})

        if(dataset_id != None):
            #If it exists update it
            self.update_dataset(dataframe, dataset_name, slug, icon_url, settings)
        else:
            # If not exists create a new one
            self.create_dataset(dataframe, dataset_name, slug, icon_url, settings)


    def delete_dataset(self, slug='', dataset_id=''):

        if slug != '':
            self._get_info_dataset_by_slug(slug)
            self._delete_dataset(self._dataset_id)
            logging.info("---- dataset '{}' deleted ----- ".format(slug))
        elif dataset_id != '' and slug == '':
            self._delete_dataset(dataset_id)
            logging.info("---- dataset '{}' deleted ----- ".format(dataset_id))
        else:
            raise Exception('Please give either a dataset id or a dataset slug')

    def create_datacard(self, channel_slug: str, title: str, search: str = "", slug: str = None, skin_code:str = None, replace: bool = False):

        if (replace == True and slug != None):
            try:
                datacard = self.get_datacard(slug)
                datacard.delete()
            except:
                pass
        elif (replace == True and slug == None):
            print("Please specify a datacard slug to replace!")
            return

        channel = self.get_channel(channel_slug)
        if channel is not None:
            channel_id = channel["id"]
        else:
            channel_id = self.create_channel(channel_slug)

        body = {
            "agentId": self._agentId,
            "channelId": channel_id,
            "name": title,
            "slug": slug
        }

        if skin_code is not None:
            body["ui"]= {"skin" : skin_code}

        if self._env == 'dev':
            smart_insight_url = url_list['BASE_URL_INSIGHT_DEV']
        if self._env == 'qa':
            smart_insight_url = url_list['BASE_URL_INSIGHT_QA']
        if self._env == 'prod':
            smart_insight_url = url_list['BASE_URL_INSIGHT_PROD']

        url = smart_insight_url + '/definitions'

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        response = s.post(url=url, json=body, headers=headers)
        response.raise_for_status()
        definition = response.json()

        if (search != ""):
            body_query = {"nl": search, "language": "en"}

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            query_url = smart_insight_url + '/definitions/' + definition["id"] + '/nl_queries/' + \
                        definition["components"][0]["id"] + '/nl'
            logging.info("QUERY URL {}".format(query_url))
            r = s.put(url=query_url, json=body_query, headers=headers)

        logging.info("DATACARD {} CREATED".format(slug))

        return Insight_Definition(self._env, self._token, self._slug, definition)

    def get_datacard(self, slug):

        url = self.smart_insight_url + "/definitions/agent/" + self._agentId + "/slug/" + slug

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        r = s.get(url=url, headers=self._headers)
        r.raise_for_status()

        definition = r.json()
        return Insight_Definition(self._env, self._token, self._slug, definition)

    def create_channel(self, name, icon='https://storage.googleapis.com/askdata/smartfeed/icons/Channel@2x.png',
                       visibility='PRIVATE', autofollow=True):

        data = {
            "userId": self.userid,
            "name": name,
            "icon": icon,
            "agentId": self._agentId,
            "visibility": visibility,
            "code": name,
            "autofollow": autofollow
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_ch + '/channels'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()
        return r.json()['id']

    def delete_channel(self, channel_slug):
        channel_id = self.get_channel(channel_slug)["id"]
        url = "{}/smartfeed/channels/{}".format(self._base_url_askdata, channel_id)

        r = requests.delete(url=url, headers=self._headers)
        r.raise_for_status()
        return r

    def execute_dataset_sync(self, dataset_slug):

        dataset_id = self.get_dataset_by_slug(self._slug, dataset_slug)

        dataset_url = self._base_url_askdata + '/smartdataset/datasets/' + dataset_id + '/sync'
        r = requests.post(url=dataset_url, headers=self._headers)
        r.raise_for_status()
        return r


    def __get_user_by_email(self, user_email, multiple=False):

        url = self._base_url_askdata + "/smartbot/share/principals?q=" + user_email
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        r = s.get(url=url, headers=self._headers)
        r.raise_for_status()
        if (r.json() != []):
            if(multiple):
                user_ids = []
                for user in r.json()[0]:
                    user_ids.append(user["id"])
                return user_ids
            else:
                user_id = r.json()[0]["id"]
                return user_id
        else:
            print("Please insert valid email")
            return None

    '''
    role: VIEW/EDIT
    '''
    def add_user_to_workspace(self, user_email, role="VIEW", add_all=False):
        user_id = self.__get_user_by_email(user_email, add_all)

        data = {
            "principals": [
                {
                    "id": user_id,
                    "type": "USER"
                }
            ],
            "permission": role
        }

        url = self._base_url_askdata+"/smartbot/agents/{}/share/entries".format(self._agentId)
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        r = s.post(url=url, headers=self._headers, json=data)
        r.raise_for_status()


    '''
    role: FOLLOWER/ADMIN
    '''
    def add_user_to_channel(self, channel_slug, user_email, role="FOLLOWER", add_all=False):

        user_id = self.__get_user_by_email(user_email, add_all)

        channel_id = self.get_channel(channel_slug)["id"]
        try:
            self.__add_user_to_ch(channel_id, user_id, role)
        except:
            logging.info("User {} can't be invited to channel, make sure the email is associated to an Askdata user".format(user_email))

    def __add_user_to_ch(self, channel_id, user_id, role="FOLLOWER"):

        data = {
            "userId":user_id,
            "role":role,
            "mute": "none"
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + "/smartfeed/channels/{}/users".format(channel_id)
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()


    def get_channel(self, channel_slug):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self._base_url_ch + '/channels?agentId=' + self._agentId + '&slug=' + channel_slug
        r = s.get(url=url, headers=self._headers)
        r.raise_for_status()
        if (r != None and r.json() != []):
            return r.json()[0]
        else:
            return None

    def create_qc(self, dataset_slug="", fields:list=[], filters=None, sorted_by=None,  dataset_id="", limit=50, offset=0, pivot=[]):
        """
        Pivot columns MUST be contained in field list
        """
        if(fields==[]):
            print("SELECT FIELDS")
            return

        if sorted_by is None:
            sorted_by = []

        if filters is None:
            filters = []

        if(dataset_id=="" and dataset_slug!=""):
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            url = self._base_url_askdata + '/smartbot/agents/agentslug/' + self.agent_slug + '/datasetslug/' + dataset_slug

            headers = {"Authorization": "Bearer" + " " + self._token}
            response = s.get(url=url, headers=headers)
            response.raise_for_status()
            r = response.json()

            datasetId = r["datasetId"]

        elif(dataset_id!=""):
            datasetId=dataset_id
        else:
            print("SPECIFY A DATASET ID OR SLUG")
            return

        body = {
            "agentId": self.agent_id,
            "datasetId": datasetId,
            "fields": fields,
            "sortedBy": sorted_by,
            "filters": filters,
            "limit": limit,
            "offset": offset,
            "pivot": pivot
        }

        url = self.smart_insight_url + "/composed_queries/sdk"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=url, json=body, headers=headers)
        response.raise_for_status()

        return response.json()

    def get_feeds_slugs(self):
        url = self._base_url_ch+"/channels?agentId="+self._agentId

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        feeds = response.json()

        slugs = []

        for feed in feeds:
            slugs.append(feed["slug"])

        return slugs

    def get_card_slug_from_feed(self, feed_slug):

        channel_id = self.get_channel(feed_slug)["id"]

        url = self._base_url_ch+"/channels/"+channel_id+"/messages"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        cards = response.json()

        slugs = []

        for card in cards:
            if card["id"] != "_system-highlight":
                slugs.append(card["insightSlug"])

        return slugs


    def get_dataset_slug_from_id(self, dataset_id: str) -> str:
        """
        get dataset slug by the dataset id instantiated with slug
        :param dataset_id: str
        :return: slug: str
        """

        list_dataset = self.list_datasets()

        if list_dataset[list_dataset['id'] == dataset_id].empty:
            raise Exception('The dataset with id: {} not exist'.format(dataset_id))
        else:
            slug = list_dataset[list_dataset['id'] == dataset_id].loc[:, 'slug'].item()

        return slug

    def list_datasets(self):

        if self._env == 'dev':
            self._base_url_dataset = url_list['BASE_URL_DATASET_DEV']
        elif self._env == 'qa':
            self._base_url_dataset = url_list['BASE_URL_DATASET_QA']
        elif self._env == 'prod':
            self._base_url_dataset = url_list['BASE_URL_DATASET_PROD']

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        dataset_url = self._base_url_dataset + '/datasets?agentId=' + self._agentId
        response = s.get(url=dataset_url, headers=self._headers)
        response.raise_for_status()
        r = response.json()
        r_df = pd.DataFrame(r)

        try:
            if r_df.empty:
                raise Exception('No datasets in the agent {}'.format(self._agentId))
            else:
                datasets_df = r_df.loc[:, ['id', 'domain', 'type', 'code', 'name', 'slug', 'description', 'createdBy', 'isActive',
                                     'accessType', 'icon', 'version', 'syncCount', 'visible', 'public', 'createdAt']]
        except Exception as e:
            datasets_df = r_df
            logging.info(e)

        return datasets_df


class Askdata(SignUp):
    '''
    Authentication Object
    '''

    def __init__(self, username='', password='', domainlogin='askdata', env='prod', token=''):

        with requests.Session() as s:

            self._token = token
            self._domainlogin = domainlogin.upper()
            self._env = env.lower()

            if self._env == 'dev':
                self.base_url_security = url_list['BASE_URL_SECURITY_DEV']
            elif self._env == 'qa':
                self.base_url_security = url_list['BASE_URL_SECURITY_QA']
            elif self._env == 'prod':
                self.base_url_security = url_list['BASE_URL_SECURITY_PROD']

            if self._env == 'dev':
                self.base_url = url_list['BASE_URL_ASKDATA_DEV']
            elif self._env == 'qa':
                self.base_url = url_list['BASE_URL_ASKDATA_QA']
            elif self._env == 'prod':
                self.base_url = url_list['BASE_URL_ASKDATA_PROD']

            if token == '':

                if username == '':
                    # add control email like
                    username = input('Askdata Username: ')
                if password == '':
                    password = getpass.getpass(prompt='Askdata Password: ')

                self.username = username

                data = {
                    "grant_type": "password",
                    "username": self.username,
                    "password": password
                }

                headers = {
                    "Authorization": "Basic YXNrZGF0YS1zZGs6YXNrZGF0YS1zZGs=",
                    "Content-Type": "application/x-www-form-urlencoded",
                    "cache-control": "no-cache,no-cache"
                }

                authentication_url = self.base_url_security + '/domain/' + self._domainlogin.lower() + '/oauth/token'

                # request token for the user
                r1 = s.post(url=authentication_url, data=data, headers=headers)
                r1.raise_for_status()
                self._token = r1.json()['access_token']
                self.r1 = r1

            authentication_url_userid = self.base_url_security + '/me'
            self._headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }

            # request userId of the user
            r_userid = s.get(url=authentication_url_userid, headers=self._headers)
            r_userid.raise_for_status()
            self.userid = r_userid.json()['id']
            self.username = r_userid.json()['userName']

    def agent(self, slug="", agent_id='') -> 'Agent':
        # Agent.__init__(self, self, slug=slug)
        return Agent(self, slug=slug, agent_id=agent_id)

    def get_token(self):
        return self._token

    def normalize_columns(self, df:pd.DataFrame):

        from askdata import integration as ig
        return ig.normalize_columns(df)


    def save_dataset(self, df:pd.DataFrame, workspace: str, dataset_slug="", replace=True, dataset_name="", locale="en", skip_import_values=False):

        df_copy = df.copy()
        '''for col in df_copy.columns:
            if (df_copy[col].dtypes == "datetime64[ns]"):
                df_copy[col] = df_copy[col].astype("str").replace("NaT", None)'''

        #df = df.where(pd.notnull(df), None)
        df_copy = self.normalize_columns(df)

        if "/" in workspace:
            agentSlug = workspace.split("/")[0]
            datasetSlug = workspace.split("/")[1]
        else:
            agentSlug = workspace
            if dataset_slug == "":
                datasetSlug = "parquet_" + uuid.uuid4().__str__()
            else:
                datasetSlug = dataset_slug

        if(replace):
            url = self.base_url + "/smartbot/agent-slugs/{}/dataset-slugs/{}/parquet".format(agentSlug, datasetSlug)

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Authorization": "Bearer" + " " + self._token
            }
            filename = dataset_slug + ".parquet"
            df_copy.to_parquet(filename)
            response = s.put(url=url, files={"file": open(filename, 'rb')}, headers=headers)
            response.raise_for_status()
            dataset = response.json()
            dataset_id = dataset["id"]
        else:
            url = self.base_url + "/smartbot/agent-slugs/{}/dataset-slugs/{}/parquet".format(agentSlug, datasetSlug)

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Authorization": "Bearer" + " " + self._token
            }
            filename = dataset_slug + ".parquet"
            df_copy.to_parquet(filename)
            response = s.post(url=url, files={"file": open(filename,'rb')}, headers=headers)
            response.raise_for_status()
            dataset = response.json()
            dataset_id = dataset["id"]

        '''url = self.base_url + "/smartbot/agents/{}/datasets/{}/data_table".format(agentSlug, datasetSlug)

        fields = []
        for column in df.columns:
            fields.append({"name": column})

        schema = {"fields": fields}

        body = {
            "data": df.to_dict(orient="record"),
            "schema": schema,
            "locale": locale,
            "defaultImportValues": (not skip_import_values)
        }

        if (replace):

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.put(url=url, json=body, headers=headers)
            response.raise_for_status()
            dataset = response.json()
            dataset_id = dataset["id"]
        else:
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.post(url=url, json=body, headers=headers)
            response.raise_for_status()
            dataset = response.json()
            dataset_id = dataset["id"]'''

        if (dataset_name != ""):
            url = self.base_url + "/smartbot/agents/" + dataset["agents"][0] + "/datasets/" + dataset_id + "/settings"
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.get(url=url, headers=headers)
            response.raise_for_status()
            settings = response.json()

            settings["label"] = dataset_name

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.put(url=url, json=settings, headers=headers)
            response.raise_for_status()
            # Execute sync
            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            dataset_url = self.base_url + '/smartdataset/datasets/' + dataset_id + '/sync'
            r = requests.post(url=dataset_url, headers=headers)
            r.raise_for_status()

    def get_dataset_as_df(self, workspace:str, dataset_slug:str):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self.base_url + '/smartbot/agents/agentslug/' + workspace + '/datasetslug/' + dataset_slug

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        r = response.json()

        datasetId = r["datasetId"]

        url = self.base_url + '/smartdataset/data_table/datasets/'+datasetId+'/file'

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        file = open(datasetId+".parquet", 'wb').write(response.content)
        return pd.read_parquet(datasetId+".parquet")

    def delete_agent(self, agent_slug):
        agent = self.agents_dataframe()[self.agents_dataframe()['slug'] == agent_slug.lower()]
        url = self.base_url + "/smartbot/agents/"+ agent.iloc[0]['id']
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        r = s.delete(url=url, headers=headers)
        r.raise_for_status()

    def join(self, df:pd.DataFrame, with_query:str, workspace:str, how="inner", on=None, left_on=None, right_on=None, lang="en"):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = self.base_url + '/smartbot/agents/agentslug/' + workspace

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=authentication_url, headers=headers)
        response.raise_for_status()
        r = response.json()
        agent_id = r["agentId"]

        url = self.base_url + "/smartbot/agents/"+agent_id+"/ask"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        body = {
            "message": with_query,
            "language": lang.lower(),
            "userId": self.userid
        }

        response = s.post(url=url, json=body, headers=self._headers)
        response.raise_for_status()
        result = response.json()["result"]
        df_query = pd.DataFrame(data=result[0]["attachment"]["body"][0]["details"]["rows"], columns=result[0]["attachment"]["body"][0]["details"]["columns"])

        if (on == None and (left_on == None and right_on == None)):
            join_conditions = self.__get_join_info(df, df_query)
            left_conditions = []
            right_conditions = []
            for condition in join_conditions:
                left_conditions.append(condition["left_on"])
                right_conditions.append(condition["right_on"])

            return df.merge(df_query, how, None, left_conditions, right_conditions)

        return df.merge(df_query, how, on, left_on, right_on)

    def load_dataset(self, workspace_slug, dataset_slug, logs=True):
        all_fields = list(self.get_columns(workspace=workspace_slug, dataset=dataset_slug)["code"])
        limit = 1000
        df = self.query(workspace=workspace_slug, dataset=dataset_slug, fields=all_fields, limit=limit)
        if(len(df.index) < 1000):
            print("Downloaded {} rows".format(len(df.index)))
            return df
        else:
            final_df = df
            while(len(df.index) == 1000):
                df = self.query(workspace=workspace_slug, dataset=dataset_slug, fields=all_fields, limit=1000, offset=limit)
                final_df = final_df.append(df)
                if(limit%10000 == 0 and logs):
                    print("Downloaded {} rows".format(limit))

                limit+=1000
            print("Downloaded {} rows".format(len(final_df.index)))
            return final_df

    def __retrieve_og_columns(self, conditions:list, df:pd.DataFrame):
        formatted_conditions = []
        for condition in conditions:
            for column in df.columns:
                if condition in column.lower():
                    formatted_conditions.append(column)

        return formatted_conditions

    def get_join_info(self, df1: pd.DataFrame, df2: pd.DataFrame):
        return self.__get_join_info(df1, df2)

    def __get_join_info(self, df1:pd.DataFrame, df2:pd.DataFrame):

        columns = []
        for field in json.loads(df1.to_json(orient="table", index=False))["schema"]["fields"]:
            field["label"] = field.pop("name")
            field["type"] = self.__choose_type(field.pop("type"))
            columns.append(field)


        columns2 = []
        for field in json.loads(df2.to_json(orient="table", index=False))["schema"]["fields"]:
            field["label"] = field.pop("name")
            field["type"] = self.__choose_type(field.pop("type"))
            columns2.append(field)


        table_left = {
            "columns": columns,
            "values": json.loads(df1.to_json(orient="values"))[:5]
        }
        table_right = {
            "columns": columns2,
            "values": json.loads(df2.to_json(orient="values"))[:5]
        }
        tables = {
            "tables": {
                "table_left": table_left,
                "table_right": table_right
            }
        }
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self.base_url + '/cfc/smartjoin'

        r = s.post(url=url, headers=self._headers, json=tables)
        r.raise_for_status()

        try:
            return r.json()["join_list"]
        except:
            logging.error(r.json())

    def __choose_type(self, var):
        if "int" in var or "num" in var or "real" in var or "double" in var or "decimal" in var or "float" in var:
            return "numeric"
        elif "text" in var or "varchar" in var or "char" in var or "default" in var:
            return "string"
        elif "date" in var or "year" in var:
            return "datetime"
        elif "time" in var:
            return "timestamp"
        elif "bool" in var or "bit" in var:
            return "bool"
        elif "blob" in var:
            return "blob"
        else:
            return var

    def get_columns(self, dataset, workspace=''):

        if(workspace==''):
            if('/' in dataset):
                workspace=dataset.split("/")[0]
                slug = dataset.split("/")[1]
        else:
            slug=dataset

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self.base_url + '/smartbot/agents/agentslug/' + workspace + '/datasetslug/' + slug

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        r = response.json()
        agent_id = r["agentId"]
        dataset_id = r["datasetId"]

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self.base_url + '/smartdataset/datasets/' + dataset_id + '/measures'

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        measures = response.json()
        columns=[]
        for measure in measures:
            columns.append({
                "code": measure["code"],
                "name": measure["name"],
                "type": "MEASURE",
                "internalType": measure["schemaMetaData"]["dataType"]
            })
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = self.base_url + '/smartdataset/datasets/' + dataset_id + '/entityTypes'

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        entityTypes = response.json()

        for entityType in entityTypes:
            columns.append({
                "code": entityType["code"],
                "name": entityType["name"],
                "type": "ENTITY_TYPE",
                "internalType": entityType["schemaMetaData"]["dataType"]
            })
        return pd.DataFrame(data=columns)

    def query(self, workspace='', dataset='', fields:list=[], filters=None, sorted_by=None, limit=50, offset=0, pivot=[], get_raw_data=False):

        if (workspace == '' and '/' in dataset):
            workspace = dataset.split("/")[0]
            dataset = dataset.split("/")[1]

        if(get_raw_data):
            return self.get("", workspace, dataset, fields=fields, filters=filters, sorted_by=sorted_by, limit=limit, offset=offset, pivot=pivot)
        else:
            agent = self.agent(slug=workspace)
            return agent.query(dataset, fields, filters, sorted_by, limit=limit, offset=offset, pivot=pivot)

    def get(self, query, workspace='', dataset='', lang="en", fields:list=[], filters=None, sorted_by=None, limit=50, offset=0, pivot=[], get_raw_data=False):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        if(workspace == '' and dataset==''):
            logging.error("Please insert a dataset or a workspace")
        elif(workspace != '' and dataset==''):
            dataset_id = ''
            try:
                authentication_url = self.base_url+ '/smartbot/agents/agentslug/' + workspace

                headers = {"Authorization": "Bearer" + " " + self._token}
                response = s.get(url=authentication_url, headers=headers)
                response.raise_for_status()
                r = response.json()
            except Exception as e:
                logging.error("{}".format(e.__str__()))
                logging.error("{} workspace not found".format(workspace))
                return None
        elif(workspace != '' and dataset!=''):
            # Get agent_id and dataset_id
            try:
                slug=dataset

                authentication_url = self.base_url+'/smartbot/agents/agentslug/'+workspace + '/datasetslug/' + slug

                headers = {"Authorization": "Bearer" + " " + self._token}
                response = s.get(url=authentication_url, headers=headers)
                response.raise_for_status()
                r = response.json()
                dataset_id = r["datasetId"]
            except Exception as e:
                logging.error("{}".format(e.__str__()))
                logging.error("{} workspace not found".format(workspace))
                return None
        else:
            # Get agent_id and dataset_id
            try:
                workspace=dataset.split("/")[0]
                slug=dataset.split("/")[1]

                authentication_url = self.base_url+'/smartbot/agents/agentslug/'+workspace + '/datasetslug/' + slug

                headers = {"Authorization": "Bearer" + " " + self._token}
                response = s.get(url=authentication_url, headers=headers)
                response.raise_for_status()
                r = response.json()
                dataset_id = r["datasetId"]
            except Exception as e:
                logging.error("{}".format(e.__str__()))
                logging.error("{} workspace not found".format(workspace))
                return None

        agent = self.agent(workspace)
        if(fields!=[]):
            qc = agent.create_qc(dataset, fields, filters, sorted_by, limit, offset, pivot)

        url = self.base_url + "/smartbot/agent-slugs/"+workspace+"/v2/ask"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        body = {
            "message": query,
            "language": lang.lower(),
            "metadata":[],
            "getRawData": get_raw_data
        }

        if(fields!=[]):
            body["metadata"].append({"qc": qc})

        response = s.post(url=url, json=body, headers=self._headers)
        response.raise_for_status()
        result = response.json()["cards"]
        try:
            if(dataset_id!=''):
                filtered_result = []
                for r in result:
                    if(r["attachment"]["body"][0]["details"]["datasetId"]==dataset_id):
                        filtered_result.append(r)
            else:
                filtered_result = result

            if self._env == 'dev':
                app_url = url_list['BASE_URL_APP_DEV']
            elif self._env == 'qa':
                app_url = url_list['BASE_URL_APP_QA']
            elif self._env == 'prod':
                app_url = url_list['BASE_URL_APP_PROD']

            url_1 = app_url + "/" + workspace+"/"+filtered_result[0]["attachment"]["body"][0]["details"]["datasetSlug"]
            display(HTML('<p>{} from the dataset: {} / {} : (<a href="{}" target="_blank">{}</a>) <- Use this source</p>'.format(self.extract_fields(filtered_result[0]["attachment"]["body"][0]["details"]["columns"]), workspace, filtered_result[0]["attachment"]["body"][0]["details"]["datasetName"], url_1, url_1)))
            if(len(filtered_result[0]["attachment"]["body"])>1):
                display(HTML("<p>Filtered by: {}</p>".format(filtered_result[0]["attachment"]["body"][1]["details"]["text"])))
            if(len(filtered_result)>1):
                display(HTML("<p>{} Alternatives:</p>".format(len(filtered_result)-1)))
                for r in filtered_result[1:]:
                    url_i = app_url + "/" + workspace + "/" + r["attachment"]["body"][0]["details"]["datasetSlug"]
                    display(HTML('<p>     {} from the dataset: {} / {} ((<a href="{}" target="_blank">{}</a>)) <- Use this source</p>'.format(self.extract_fields(r["attachment"]["body"][0]["details"]["columns"]), workspace, r["attachment"]["body"][0]["details"]["datasetName"], url_i, url_i)))
            if get_raw_data:
                df = pd.DataFrame(data=filtered_result[0]["attachment"]["body"][0]["details"]["rawData"], columns=filtered_result[0]["attachment"]["body"][0]["details"]["columns"])
            else:
                df = pd.DataFrame(data=filtered_result[0]["attachment"]["body"][0]["details"]["rows"], columns=filtered_result[0]["attachment"]["body"][0]["details"]["columns"])
            return df
        except Exception as e:
            logging.error("Query did not produce any result")
            logging.error(e.__str__())

    def load_agents(self):

        if self._env == 'dev':
            authentication_url = url_list['BASE_URL_AGENT_DEV']

        if self._env == 'qa':
            authentication_url = url_list['BASE_URL_AGENT_QA']

        if self._env == 'prod':
            authentication_url = url_list['BASE_URL_AGENT_PROD']

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        # request of all agents of the user/token
        response = s.get(url=authentication_url, headers=self._headers)
        response.raise_for_status()

        return response.json()

    def agents_dataframe(self):
        return pd.DataFrame(self.load_agents())

    def signup_user(self, username, password, firstname='-', secondname='-', title='-'):
        response = super().signup_user(username, password, firstname, secondname, title)
        return response


    def create_agent(self, agent_name):

        data = {
            "name": agent_name,
            "language": "en"
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = self.base_url + '/smartbot/agents'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()

        return r

    def extract_fields(self, fields:list)->str:
        column_names = ""
        for field in fields:
            column_names+=(field+ " , ")
        return column_names[:-3]



from askdata.smartgraph import smart_opendata, opendata_specific_result


if __name__ == "__main__":
    nl = "wine consumption by countries"
    response = smart_opendata(sentence=nl)
    extraction = opendata_specific_result(response=response, request="top1_df", show=True)
    print(extraction)



from askdata import human2query
from askdata.smartquery import SmartQuery, Query, Field, Condition, Sorting, SQLSorting, SQLOperator, TimeOperator, \
    BooleanOperator, CompositeCondition

if __name__ == "__main__":

    # Query2SQL
    smartquery = SmartQuery(
        queries=[
            Query(
                fields=[
                    Field(aggregation="SUM", column="incidents", alias="sum_incidents",
                          internalDataType="NUMERIC",
                          sourceDataType="INT64"),
                    Field(column="customer_name", alias="Customer",
                          internalDataType="STRING",
                          sourceDataType="VARCHAR"),
                    Field(aggregation="YEAR", column="acquired", alias="Acquired Date",
                          internalDataType="DATE",
                          sourceDataType="DATE")
                ],
                where=[
                    CompositeCondition(conditions=[
                        Condition(field=Field(column="customer_name", alias="Customer",
                                              internalDataType="STRING",
                                              sourceDataType="VARCHAR"), operator=SQLOperator.IN,
                                  value=["Franceaco Doe", "Claude Rudolf"]),
                        Condition(field=Field(aggregation="YEAR", column="acquired", alias="Acquired Date",
                                              internalDataType="DATE",
                                              sourceDataType="DATE"), operator=TimeOperator.RANGE,
                                  value=["2018-01-01", "2019-12-31"])
                    ], operator=BooleanOperator.AND)

                ],
                orderBy=[
                    Sorting(field="Acquired Date", order=SQLSorting.DESC)
                ],
                limit=6
            )
        ]
    )

    response = human2query.query2sql(smartquery=smartquery, driver="MySQL")
    print(response)
    response = human2query.complex_field_calculator(smartquery=smartquery, driver="MySQL")
    print(response)
    response = human2query.complex_filter_calculator(smartquery=smartquery, driver="MySQL")
    print(response)



import os
import pandas as pd
import numpy as np
import sys
from datetime import date as dt
from datetime import timedelta as td 
import time
from time import sleep
import datetime
from dateutil import parser
import pytz
import requests
import math



"""

MANDATORY INPUT: token, api_key

OPTIONAL INPUT:

- start: n days ago for starting the date range.
        Default start = 7 days ago.
        
- end: n days ago for ending the date range.
        Default end = 0 days ago.


"""




def get_aircall_calls(token, api_key, #mandatory
                      
                      start = 7, end = 0 #optional
                     ):

    
    
    start_date = (dt.today()-td(days = start)).isoformat()
    # converted to UNIX timefromat 
    start_date = math.trunc(time.mktime(datetime.datetime.strptime(start_date, "%Y-%m-%d").timetuple()))

    end_date = (dt.today()-td(days = end)).isoformat()
    # converted to UNIX timefromat 
    end_date = math.trunc(time.mktime(datetime.datetime.strptime(end_date, "%Y-%m-%d").timetuple()))
    
    link = f'https://{token}:{api_key}@api.aircall.io/v1/calls?from={start_date}&to={end_date}&page=1&per_page=50'

    headers = {'accept': 'application/json'}

    response = requests.request("GET", link, headers=headers)

    dati = response.json()

    total_items = dati["meta"]["total"]
    
    
    records = []

    if total_items >= 10000:

        print("Range interval reduce as:")


        day_intervals = []

        while True:

            left = start

            right = max(end, start - 11)

            day_intervals.append((left, right))

            start = right + 1

            if right == end:

                break




        for start, end in day_intervals:

            print("start {}, end {}".format(start, end))

            start_date = (dt.today()-td(days = start)).isoformat()
                # converted to UNIX timefromat 
            start_date =math.trunc(time.mktime(datetime.datetime.strptime(start_date, "%Y-%m-%d").timetuple()))

            end_date = (dt.today()-td(days = end)).isoformat()
            # converted to UNIX timefromat 
            end_date = math.trunc(time.mktime(datetime.datetime.strptime(end_date, "%Y-%m-%d").timetuple()))
            
            new_link = f'https://{token}:{api_key}@api.aircall.io/v1/calls?from={start_date}&to={end_date}&page=1&per_page=50'
    
            new_headers = {'accept': 'application/json'}

            new_response = requests.request("GET", new_link, headers=new_headers)

            new_dati = new_response.json()

            new_total_items = new_dati["meta"]["total"]
            
            if new_total_items >= 10000:
                
                print("Second range interval reduced as:")
                
             

                for new_left in range(start, end, -1):
                    
                    print("start {}, end {}".format(new_left, new_left - 1))

                    new_start_date = (dt.today() - td(days = new_left)).isoformat()
                    
                    # converted to UNIX timefromat
                    
                    new_start_date = math.trunc(time.mktime(datetime.datetime.strptime(new_start_date, "%Y-%m-%d").timetuple()))
                    
                    new_end_date = (dt.today() - td(days = new_left - 1)).isoformat()
                    
                    new_end_date = math.trunc(time.mktime(datetime.datetime.strptime(new_end_date, "%Y-%m-%d").timetuple()))
                    
                    records += run_all_requests(new_start_date, new_end_date, token, api_key)
             
           
            
            else:


                records += run_all_requests(start_date, end_date, token, api_key)

    else:

        records += run_all_requests(start_date, end_date, token, api_key)
    
    aircall_calls = pd.DataFrame(records)


    unix_to_readable = ["answered_at", "ended_at", "started_at"]

    for unix_col in unix_to_readable:

        #aircall_calls[unix_col] = pd.to_datetime(aircall_calls[unix_col], unit='s')
        
        #convert to italian timezone
        aircall_calls[unix_col] = pd.to_datetime(aircall_calls[unix_col], unit='s', utc=True, ).dt.tz_convert('Europe/Rome').dt.tz_localize(None)

    
    return aircall_calls




def run_all_requests(start_date, end_date, token, api_key):
    
    
    
    records = []
    
    page = 1
    
    run = True
    
    while run:

        sleep(1)

        


        link = f'https://{token}:{api_key}@api.aircall.io/v1/calls?from={start_date}&to={end_date}&page={page}&per_page=50'

        headers = {'accept': 'application/json'}

        response = requests.request("GET", link, headers=headers)

        dati = response.json()

        total_pages = math.ceil(dati["meta"]["total"]/50)

        if page == 1:
            total_items = dati["meta"]["total"]



            print("Total number of pages", total_pages)

            print("Total number of items", total_items)

        '''

        if page%30 == 0: 
            
            print("---------")
            print("Page N.", page)

            print(dati["meta"])

        '''
        
        
        if dati["meta"]["count"] == 0:
            run = False


        else:
            page += 1

            for d in dati['calls']:

                try:
                    id = d['id']
                except:
                    id  = ''
                try:
                    cost_cent = d['cost']
                except:
                    cost_cent  = ''                
                try:
                    direction = d['direction']
                except:
                    direction  = ''
                try:
                    status = d['status']
                except:
                    status  = ''
                try:
                    started_at = d['started_at']
                except:
                    started_at  = ''
                try:
                    answered_at = d['answered_at']
                except:
                    answered_at  = ''
                try:
                    ended_at = d['ended_at']
                except:
                    ended_at  = ''
                try:
                    duration = d['duration']
                except:
                    duration  = ''
                try:
                    raw_digits = d['raw_digits']
                except:
                    raw_digits  = ''
                try:
                    user_id = d['user'][0]['id']
                except:
                    try:
                        user_id  = d['user']['id']
                    except:

                        user_id = ''
                try:
                    user_name = d['user'][0]['name']
                except:
                    try:
                        user_name  = d['user']['name']
                    except:
                        user_name = ''
                try:
                    contact_id = d['contact']['id']
                except:
                    contact_id  = ''
                try:
                    contact_name = d['contact']['first_name']
                except:
                    contact_name  = ''
                try:
                    contact_last_name = d['contact']['last_name']
                except:
                    contact_last_name  = ''
                try:
                    cnt_company_name = d['contact']['company_name']
                except:
                    cnt_company_name  = ''

                records.append({
                    'id' : id,
                    'direction' : direction,
                    'status' : status,
                    'started_at' : started_at, #
                    'answered_at' : answered_at, #
                    'ended_at' : ended_at,
                    'duration' : duration, #
                    'raw_digits' : raw_digits, #deal_id
                    'user_id' : user_id,
                    'user_name' : user_name,
                    'contact_id' : contact_id,
                    'contact_name' : contact_name,
                    'contact_last_name' : contact_last_name,
                    'cnt_company_name' : cnt_company_name,
                    'cost_cent' : cost_cent
                })
    
    return records



from askdata.integrations import alphavantage
import json
import dataset
import pandas as pd

""" QUERY """

class Alphavantage(dataset.Dataset):
    
    dataset_settings = """{
            "payload": {
                "page": 0,
                "totalElements": 8,
                "data": [
                    {
                        "datasetId": "f324ef9a-279b-4105-901b-c47c624edf7f-DATA_TABLE-0d129d26-58e5-441f-b9b3-8c80efdeb094",
                        "schemaMetaData": {
                            "columnId": "area_geografica",
                            "columnName": "area_geografica",
                            "dataType": "VARCHAR",
                            "dataExample": "Calabria",
                            "internalDataType": "STRING",
                            "indexedWith": null,
                            "join": null,
                            "details": {}
                        },
                        "parameterType": "ENTITY_TYPE",
                        "code": "AREA_GEOGRAFICA",
                        "name": "Area geografica",
                        "description": null,
                        "synonyms": [
                            "area",
                            "geografia",
                            "area geografica",
                            "area geograficas",
                            "regione"
                        ],
                        "icon": "",
                        "sampleQueries": [],
                        "importValues": true,
                        "mandatory": false,
                        "enabled": true,
                        "advancedConfiguration": {},
                        "viewValues": "View",
                        "custom": false,
                        "dynamicParameterValues": [],
                        "searchable": false,
                        "nameTransformer": null,
                        "synonymTransformers": null
                    }
                ]
            }
        }"""

    def discovery(self):
        return self.dataset_settings



from setuptools import setup, find_packages

setup(name='askdata',

      version='1.3.98',

      description='The official Askdata Python SDK',
      url='https://github.com/AskdataInc/askdata-api-python-client',
      author=['Giuseppe De Maio', 'Matteo Giacalone', 'Luca Sarcona', 'Simone Di Somma', 'Massimiliano Bruni',
              'Oscar Caruso', 'Gennaro Vaccaro', 'Maria Luisa Croci'],
      author_email='datascience@askdata.com',
      license='Apache License 2.0',
      packages=find_packages(exclude=["dev", "*.tests", "*.tests.*", "tests.*", "tests", "entity.py", "feed.py"]),
      long_description=open('README.md').read(),
      long_description_content_type="text/markdown",
      include_package_data=True,
      install_requires=[
          'pandas>=1.1.2',
          'numpy>=1.19.2',
          'PyYAML>=5.1',
          'yaml-1.3',
          'requests>=2',
          'urllib3>=1',
          'sqlalchemy>=1.3.8',
          'mysql-connector>=2.2.9',
          'NotebookScripter==6.0.0',
          'jsons',
          'googleads>=31.0.0',
          'google-api-python-client>=2.42.0',
          'oauthlib>=3.2.0',
          'oauth2client==4.1.3',
          'facebook-business>=12.0.1',
          'hubspot3>=3.2.51'
      ],

      classifiers=[
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent",
          "License :: OSI Approved :: MIT License",
          "Development Status :: 3 - Alpha",
      ],
      keywords='nlp',
      python_requires='>=3, <4',
      zip_safe=False,
      setup_requires=['nose'],
      test_suite='nose.collector'
      )



#required !pip install googleads -q


import pandas as pd
import numpy as np
import datetime
import io
import os
import glob
import sys
from googleads import adwords
from datetime import datetime, timedelta


import _locale


"""
MANDATORY INPUT:

start_date, end_date as string "yyyy-mm-dd"

OPTIONAL INPUTS: 



- yaml_file_path: the path of 'googleads.yaml'. 
                Optional, but mandatory to have it in same folder script if used the default parameter.

"""


def get_google_ads(start_date, end_date, admin_account,
                   
                   
                   ## optional
                   
                   yaml_file_path = 'googleads.yaml'
                  ):

    _locale._getdefaultlocale = (lambda *args: ['en_US', 'UTF-8'])

    



    def run_Hour0fDay_kip_report(key, acc_id, start_date, end_date):
    # Define output as a string
        output= io.StringIO()
        adwords_client = adwords.AdWordsClient.LoadFromStorage(yaml_file_path)
        adwords_client.SetClientCustomerId(acc_id)    
        report_downloader = adwords_client.GetReportDownloader(version='v201809')
        report_query = (adwords.ReportQueryBuilder()
                          .Select(
                              'Month'
                              ,'Date'
                            #   ,'AccountId'
                              ,'CampaignId'
                              ,'ExternalCustomerId'
                              ,'CampaignName'
                              ,'CampaignStatus'
                            #   ,'CampaignType'
                              ,'Amount'
                              ,'AccountCurrencyCode'
                              ,'Clicks'
                              ,'Impressions'
                              ,'Ctr'
                              ,'AverageCpc'
                              ,'Cost'
                              ,'Conversions'
                              ,'ViewThroughConversions'
                              ,'CostPerConversion'
                              ,'ConversionRate'
                              ,'AverageCpm'
                              )
                          .From('CAMPAIGN_PERFORMANCE_REPORT')
                        #   .Where('CampaignStatus').In('ENABLED')
                          .During(start_date+ ','+end_date) 
                          .Build())

        report_downloader.DownloadReportWithAwql(report_query, 'CSV', output, skip_report_header=True,
                  skip_column_header=False, skip_report_summary=True,
                  include_zero_impressions=False)

        output.seek(0)

        types= { 'CampaignId':pd.np.int64, 'Clicks': pd.np.float64, 'Impressions': pd.np.float64,
                 'Cost': pd.np.float64,'Conversions': pd.np.float64,'ConversionValue': pd.np.float64  }

        df = pd.read_csv(output,low_memory=False, dtype= types, na_values=[' --'])
        # delete the first and last column
        df['Brand']=key
        # micro amount 1000000
        df['Cost']=df.Cost/1000000
        print(df.head())
        return df


    google_ads = pd.DataFrame(columns = ['Month', 'Day', 'Campaign ID', 'Customer ID', 'Campaign',
           'Campaign state', 'Budget', 'Currency', 'Clicks', 'Impressions', 'CTR',
           'Avg. CPC', 'Cost', 'Conversions', 'View-through conv.', 'Cost / conv.',
           'Conv. rate', 'Avg. CPM', 'Brand'])



    for k, v in admin_account.items():

        df=run_Hour0fDay_kip_report(k, v,start_date, end_date)

        google_ads = pd.concat([google_ads, df])
        
        
    return google_ads


from askdata.human2query import nl2query

if __name__ == "__main__":
    nl = "{{measures.measureA.name}} nel {{entityTypes.entityTypeA.value}}"
    smartquery, version = nl2query(nl, language="en-US")
    print(smartquery)
    print(version)



import json
import jsons
import operator
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import random
import logging
import requests
import pandas as pd
import os
import yaml
import warnings

warnings.simplefilter(action='ignore', category=FutureWarning)

ops = {
    '<': operator.lt,
    '<=': operator.le,
    '==': operator.eq,
    '!=': operator.ne,
    '>=': operator.ge,
    '>': operator.gt
}

root_dir = os.path.abspath(os.path.dirname(__file__))
# retrieving base url
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to Python the dictionary format
    url_list = yaml.load(file, Loader=yaml.FullLoader)


def ask_dataframe(dataframe: pd.DataFrame, query: str):
    human2sql_request = {
        "dataframe": dataframe.to_dict(),
        "query": query
    }

    human2sql_url = url_list['BASE_URL_HUMAN2SQL_DEV']

    human2sql_response = requests.post(human2sql_url, json=human2sql_request)
    response_df = []
    if human2sql_response.ok:
        res = human2sql_response.json()
        if 'result' in res:
            all_dfs = res['result']
            for df in all_dfs:
                response_df.append(pd.DataFrame(df))
        if 'messages' in res and res['messages']:
            for mex in res['messages']:
                print(mex)
    else:
        print("Error: " + str(human2sql_response))

    return response_df


def get_conditional_phrases(conditions, phrase1, phrase2):
    bool_array = []
    for condition in conditions:
        for op in ops:
            if op in condition:
                splitted = condition.split(' ' + op + ' ')
                operator = ops[op]
                bool_array.append(operator(splitted[0], splitted[1]))
        # bool_array.append(eval(condition))
    if False in bool_array:
        return phrase2
    else:
        return phrase1


def get_random_synonymous(synonyms):
    random.seed()
    nKeys = len(synonyms.keys())
    pickedItem = random.randint(0, nKeys - 1)
    synPicked = synonyms[pickedItem]
    return synPicked


def words_to_digits(phrase):
    words = phrase.split(" ")
    num = []
    for word in words:
        if "." in word:
            word = word.replace(".", "")
        if "," in word:
            word = word.replace(",", "")
        if word.isdigit():
            num.append(int(word))
    return num


def add_random_synonymous_to_sentence(phrase, placeholder, synonyms):
    random.seed()
    nKeys = len(synonyms.keys())
    pickedItem = random.randint(0, nKeys - 1)
    synPicked = synonyms[pickedItem]
    phrase = phrase.replace(placeholder, synPicked)
    return phrase


def data2nl(df):
    # The parameter "df" is a dataframe, but it will be converted into a dict, and called df in each case

    if not df.empty:

        response_df = pd.DataFrame()

        headers = {
            "Content-Type": "application/json"
        }

        for i, row in df.iterrows():

            data = {
                "data": str(row)
            }

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            url = url_list['BASE_URL_DATA2NL_DEV'] + "/data2nl"

            r = s.post(url=url, headers=headers, json=data)
            r.raise_for_status()

            try:
                nl = r.json()['nl']
                row['nl'] = nl
            except Exception as e:
                logging.error(str(e))
                row['generated_nl'] = ""
            tmp_df = pd.DataFrame()
            tmp_df = tmp_df.append(row)
            response_df = pd.concat([response_df, tmp_df], ignore_index=True, axis=0)
        return response_df
    else:
        print("Input DataFrame is empty!")
        df_response = pd.DataFrame()
        return df_response


def query2sql(smartquery, driver):
    # Google Pod
    headers = {
        "Content-Type": "application/json"
    }

    s = requests.Session()
    s.keep_alive = False
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    s.mount('https://', HTTPAdapter(max_retries=retries))

    url = url_list['BASE_URL_QUERY2SQL_DEV'] + "/query_to_sql"

    stringed_smartquery = jsons.dumps(smartquery, strip_nulls=True)
    smartquery = json.loads(stringed_smartquery)

    data = {
        "smartquery": smartquery,
        "driver": driver
    }

    r = s.post(url=url, headers=headers, json=data)
    r.raise_for_status()

    try:
        dict_response = r.json()
        sql = dict_response['sql']
        return sql
    except Exception as e:
        logging.error(str(e))


def nl2query(nl, language="en-US"):
    # Google Pod
    headers = {
        "Content-Type": "application/json"
    }

    s = requests.Session()
    s.keep_alive = False
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    s.mount('https://', HTTPAdapter(max_retries=retries))

    url = url_list['BASE_URL_NL2QUERY_DEV'] + "/query"

    if "it" in language:
        language = "it-IT"
    else:
        language = "en-US"

    data = {
        "nl_ner": nl,
        "lang": language
    }

    r = s.post(url=url, headers=headers, json=data)
    r.raise_for_status()

    try:
        response = r.json()
        smartquery, version = response['smartquery'], response['model_version']
        return smartquery, version
    except Exception as e:
        logging.error(str(e))


def complex_field_calculator(smartquery, driver):
    # Google Pod
    headers = {
        "Content-Type": "application/json"
    }

    s = requests.Session()
    s.keep_alive = False
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    s.mount('https://', HTTPAdapter(max_retries=retries))

    url = url_list['BASE_URL_QUERY2SQL_DEV'] + "/query_to_sql"

    stringed_smartquery = jsons.dumps(smartquery, strip_nulls=True)
    smartquery = json.loads(stringed_smartquery)

    data = {
        "smartquery": smartquery,
        "driver": driver
    }

    r = s.post(url=url, headers=headers, json=data)
    r.raise_for_status()

    try:
        dict_response = r.json()
        sql = dict_response['sql']
        return sql
    except Exception as e:
        logging.error(str(e))


def complex_filter_calculator(smartquery, driver):
    # Google Pod
    headers = {
        "Content-Type": "application/json"
    }

    s = requests.Session()
    s.keep_alive = False
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    s.mount('https://', HTTPAdapter(max_retries=retries))

    url = url_list['BASE_URL_QUERY2SQL_DEV'] + "/query_to_sql"

    stringed_smartquery = jsons.dumps(smartquery, strip_nulls=True)
    smartquery = json.loads(stringed_smartquery)

    data = {
        "smartquery": smartquery,
        "driver": driver
    }

    r = s.post(url=url, headers=headers, json=data)
    r.raise_for_status()

    try:
        dict_response = r.json()
        sql = dict_response['sql']
        return sql
    except Exception as e:
        logging.error(str(e))
